from .knapsack import Knapsack, SlackFreeKnapsack
from .maximumcut import MaximumCut
from .minimumvertexcover import MinimumVertexCover
from .numberpartition import NumberPartition
from .shortestpath import ShortestPath
from .tsp import TSP
from .portfoliooptimization import PortfolioOptimization
from .maximalindependentset import MIS
from .binpacking import BinPacking
from .vehiclerouting import VRP
from .converters import FromDocplex2IsingModel
from .qubo import QUBO



from typing import List
from copy import deepcopy

from .gatemap import RotationGateMapFactory, GateMap
from .gatemaplabel import GateMapType
from .operators import Hamiltonian


class HamiltonianMapper(object):
    def generate_gate_maps(
        hamil_obj: Hamiltonian, gatemap_type: GateMapType
    ) -> List[GateMap]:
        """
        This method gets the rotation gates based on the input Hamiltonian into the Mapper

        Parameters
        ----------
        hamil_obj : `Hamiltonian`
            The Hamiltonian object to construct the gates from
        input_label : `GateMapType`
            Input label defining the type of gate

        Returns
        -------
        `list[GateMap]`
            List of RotationGateMap objects defining part of the circuit
        """
        assert isinstance(
            gatemap_type, GateMapType
        ), f"gatemap_type must be of supported types: {GateMapType.supported_types}"
        return RotationGateMapFactory.rotationgatemap_list_from_hamiltonian(
            hamil_obj, gatemap_type
        )

    def repeat_gate_maps(
        gatemap_list: List[GateMap], n_layers: int
    ) -> List[List[GateMap]]:
        """
        Repeat the gates for n_layers based on the input gatelist

        Parameters
        ----------
        gatemap_list : `List[GateMap]`
            Repeat the gates from the gatemap_list
        n_layers: `int`
            The number of times the layer of gates have to be repeated.
        """
        output_gate_list = []
        for each_layer in range(n_layers):
            output_gate_list.append(
                RotationGateMapFactory.gatemaps_layer_relabel(gatemap_list, each_layer)
            )

        return output_gate_list



import unittest
import numpy as np

from qiskit import QuantumCircuit

from openqaoa.qaoa_components import (
    PauliOp,
    Hamiltonian,
    QAOADescriptor,
    QAOAVariationalExtendedParams,
    QAOAVariationalStandardParams,
)
from openqaoa_qiskit.backends import (
    QAOAQiskitBackendStatevecSimulator,
    QAOAQiskitBackendShotBasedSimulator,
)
from openqaoa.backends import QAOAvectorizedBackendSimulator
from openqaoa.utilities import X_mixer_hamiltonian, ring_of_disagrees


class TestingQAOAQiskitSimulatorBackend(unittest.TestCase):

    """This Object tests the QAOA Qiskit Simulator Backend objects, which is
    tasked with the creation and execution of a QAOA circuit for the qiskit
    library and its local backends.
    """

    def test_circuit_angle_assignment_statevec_backend(self):
        """
        A tests that checks if the circuit created by the Qiskit Backend
        has the appropriate angles assigned before the circuit is executed.
        Checks the circuit created on IBM Simulator Backends.
        """

        ntrials = 10

        nqubits = 3
        p = 2
        weights = [
            [np.random.rand(), np.random.rand(), np.random.rand()]
            for i in range(ntrials)
        ]
        init_hadamards = [np.random.choice([True, False]) for i in range(ntrials)]
        constants = [np.random.rand() for i in range(ntrials)]

        for i in range(ntrials):
            gammas = [np.random.rand() * np.pi for i in range(p)]
            betas = [np.random.rand() * np.pi for i in range(p)]

            print(gammas)
            print(betas)

            cost_hamil = Hamiltonian(
                [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
                weights[i],
                constants[i],
            )
            mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
            qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
            variate_params = QAOAVariationalStandardParams(
                qaoa_descriptor, betas, gammas
            )

            qiskit_statevec_backend = QAOAQiskitBackendStatevecSimulator(
                qaoa_descriptor, None, None, False
            )

            statevec_circuit = qiskit_statevec_backend.qaoa_circuit(variate_params)

            # Trivial Decomposition
            main_circuit = QuantumCircuit(3)
            main_circuit.rzz(2 * weights[i][0] * gammas[0], 0, 1)
            main_circuit.rzz(2 * weights[i][1] * gammas[0], 1, 2)
            main_circuit.rzz(2 * weights[i][2] * gammas[0], 0, 2)
            main_circuit.rx(-2 * betas[0], 0)
            main_circuit.rx(-2 * betas[0], 1)
            main_circuit.rx(-2 * betas[0], 2)
            main_circuit.rzz(2 * weights[i][0] * gammas[1], 0, 1)
            main_circuit.rzz(2 * weights[i][1] * gammas[1], 1, 2)
            main_circuit.rzz(2 * weights[i][2] * gammas[1], 0, 2)
            main_circuit.rx(-2 * betas[1], 0)
            main_circuit.rx(-2 * betas[1], 1)
            main_circuit.rx(-2 * betas[1], 2)

            self.assertEqual(
                main_circuit.to_instruction().definition,
                statevec_circuit.to_instruction().definition,
            )

    def test_circuit_angle_assignment_statevec_backend_w_hadamard(self):
        """
        Checks for consistent if init_hadamard is set to True.
        """

        ntrials = 10

        nqubits = 3
        p = 2
        weights = [
            [np.random.rand(), np.random.rand(), np.random.rand()]
            for i in range(ntrials)
        ]
        init_hadamards = [np.random.choice([True, False]) for i in range(ntrials)]
        constants = [np.random.rand() for i in range(ntrials)]

        for i in range(ntrials):
            gammas = [np.random.rand() * np.pi for i in range(p)]
            betas = [np.random.rand() * np.pi for i in range(p)]

            cost_hamil = Hamiltonian(
                [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
                weights[i],
                constants[i],
            )
            mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
            qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
            variate_params = QAOAVariationalStandardParams(
                qaoa_descriptor, betas, gammas
            )

            qiskit_statevec_backend = QAOAQiskitBackendStatevecSimulator(
                qaoa_descriptor, None, None, True
            )

            statevec_circuit = qiskit_statevec_backend.qaoa_circuit(variate_params)

            # Trivial Decomposition
            main_circuit = QuantumCircuit(3)
            main_circuit.h([0, 1, 2])
            main_circuit.rzz(2 * weights[i][0] * gammas[0], 0, 1)
            main_circuit.rzz(2 * weights[i][1] * gammas[0], 1, 2)
            main_circuit.rzz(2 * weights[i][2] * gammas[0], 0, 2)
            main_circuit.rx(-2 * betas[0], 0)
            main_circuit.rx(-2 * betas[0], 1)
            main_circuit.rx(-2 * betas[0], 2)
            main_circuit.rzz(2 * weights[i][0] * gammas[1], 0, 1)
            main_circuit.rzz(2 * weights[i][1] * gammas[1], 1, 2)
            main_circuit.rzz(2 * weights[i][2] * gammas[1], 0, 2)
            main_circuit.rx(-2 * betas[1], 0)
            main_circuit.rx(-2 * betas[1], 1)
            main_circuit.rx(-2 * betas[1], 2)

            self.assertEqual(
                main_circuit.to_instruction().definition,
                statevec_circuit.to_instruction().definition,
            )

    def test_prepend_circuit(self):
        """
        Checks if prepended circuit has been prepended correctly.
        """

        nqubits = 3
        p = 1
        weights = [1, 1, 1]
        gammas = [1 / 8 * np.pi]
        betas = [1 / 8 * np.pi]
        shots = 10000

        # Prepended Circuit
        prepend_circuit = QuantumCircuit(3)
        prepend_circuit.x([0, 1, 2])

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
        variate_params = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)

        qiskit_statevec_backend = QAOAQiskitBackendStatevecSimulator(
            qaoa_descriptor, prepend_circuit, None, True
        )
        qiskit_shot_backend = QAOAQiskitBackendShotBasedSimulator(
            qaoa_descriptor, shots, prepend_circuit, None, True, 1.0
        )

        statevec_circuit = qiskit_statevec_backend.qaoa_circuit(variate_params)
        shot_circuit = qiskit_shot_backend.qaoa_circuit(variate_params)
        shot_circuit.remove_final_measurements()

        # Trivial Decomposition
        main_circuit = QuantumCircuit(3)
        main_circuit.x([0, 1, 2])
        main_circuit.h([0, 1, 2])
        main_circuit.rzz(2 * gammas[0], 0, 1)
        main_circuit.rzz(2 * gammas[0], 1, 2)
        main_circuit.rzz(2 * gammas[0], 0, 2)
        main_circuit.rx(-2 * betas[0], 0)
        main_circuit.rx(-2 * betas[0], 1)
        main_circuit.rx(-2 * betas[0], 2)

        self.assertEqual(
            main_circuit.to_instruction().definition,
            statevec_circuit.to_instruction().definition,
        )
        self.assertEqual(
            main_circuit.to_instruction().definition,
            shot_circuit.to_instruction().definition,
        )

    def test_append_circuit(self):
        """
        Checks if appended circuit is appropriately appended to the back of the
        QAOA Circuit.
        """

        nqubits = 3
        p = 1
        weights = [1, 1, 1]
        gammas = [1 / 8 * np.pi]
        betas = [1 / 8 * np.pi]
        shots = 10000

        # Appended Circuit
        append_circuit = QuantumCircuit(3)
        append_circuit.x([0, 1, 2])

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
        variate_params = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)

        qiskit_statevec_backend = QAOAQiskitBackendStatevecSimulator(
            qaoa_descriptor, None, append_circuit, True
        )
        qiskit_shot_backend = QAOAQiskitBackendShotBasedSimulator(
            qaoa_descriptor, shots, None, append_circuit, True, 1.0
        )

        statevec_circuit = qiskit_statevec_backend.qaoa_circuit(variate_params)
        shot_circuit = qiskit_shot_backend.qaoa_circuit(variate_params)
        shot_circuit.remove_final_measurements()

        # Standard Decomposition
        main_circuit = QuantumCircuit(3)
        main_circuit.h([0, 1, 2])
        main_circuit.rzz(2 * gammas[0], 0, 1)
        main_circuit.rzz(2 * gammas[0], 1, 2)
        main_circuit.rzz(2 * gammas[0], 0, 2)
        main_circuit.rx(-2 * betas[0], 0)
        main_circuit.rx(-2 * betas[0], 1)
        main_circuit.rx(-2 * betas[0], 2)
        main_circuit.x([0, 1, 2])

        self.assertEqual(
            main_circuit.to_instruction().definition,
            statevec_circuit.to_instruction().definition,
        )
        self.assertEqual(
            main_circuit.to_instruction().definition,
            shot_circuit.to_instruction().definition,
        )

    def test_qaoa_circuit_wavefunction_expectation_equivalence_1(self):
        """
        The following tests with a similar naming scheme check for consistency
        between the outputs of the qiskit statevector simulator and the
        vectorized backend.
        We compare both the wavefunctions returned by the Backends and the
        expectation values.
        """

        nqubits = 3
        p = 1
        weights = [1, 2, 3]
        gammas = [[3], [2]]
        betas = [[1], [1 / 8]]

        for i in range(2):
            cost_hamil = Hamiltonian(
                [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
                weights,
                1,
            )
            mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
            qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
            variate_params = QAOAVariationalStandardParams(
                qaoa_descriptor, betas[i], gammas[i]
            )

            qiskit_backend = QAOAQiskitBackendStatevecSimulator(
                qaoa_descriptor, None, None, True
            )

            qiskit_wavefunction = qiskit_backend.wavefunction(variate_params)
            qiskit_expectation = qiskit_backend.expectation(variate_params)

            vector_backend = QAOAvectorizedBackendSimulator(
                qaoa_descriptor, None, None, True
            )
            vector_wavefunction = vector_backend.wavefunction(variate_params)
            vector_expectation = vector_backend.expectation(variate_params)

            self.assertAlmostEqual(qiskit_expectation, vector_expectation)

            for j in range(2**nqubits):
                self.assertAlmostEqual(
                    qiskit_wavefunction[j].real, vector_wavefunction[j].real
                )
                self.assertAlmostEqual(
                    qiskit_wavefunction[j].imag, vector_wavefunction[j].imag
                )

    def test_qaoa_circuit_wavefunction_expectation_equivalence_2(self):
        """Due to the difference in the constructions of the statevector simulators,
        there is a global phase difference between the results obtained from
        qiskit's statevector simulator and OpenQAOA's vectorised simulator. In order to
        show the equivalence between the wavefunctions produced, the expectation
        of a random operator is computed.
        """

        nqubits = 3
        p = 1
        weights = [1, 2, 3]
        gammas = [[1 / 8 * np.pi], [1 / 8 * np.pi]]
        betas = [[1], [1 / 8 * np.pi]]

        for i in range(2):
            cost_hamil = Hamiltonian(
                [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
                weights,
                1,
            )

            mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)

            qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
            variate_params = QAOAVariationalStandardParams(
                qaoa_descriptor, betas[i], gammas[i]
            )

            qiskit_backend = QAOAQiskitBackendStatevecSimulator(
                qaoa_descriptor, None, None, True
            )

            qiskit_wavefunction = qiskit_backend.wavefunction(variate_params)
            qiskit_expectation = qiskit_backend.expectation(variate_params)

            vector_backend = QAOAvectorizedBackendSimulator(
                qaoa_descriptor, None, None, True
            )
            vector_wavefunction = vector_backend.wavefunction(variate_params)
            vector_expectation = vector_backend.expectation(variate_params)

            random_operator = (
                np.random.rand(2**nqubits, 2**nqubits)
                + np.random.rand(2**nqubits, 2**nqubits) * 1j
            )
            random_herm = random_operator + random_operator.conj().T

            expect_qiskit = np.matmul(
                np.array(qiskit_wavefunction).T.conjugate(),
                np.matmul(random_herm, np.array(qiskit_wavefunction)),
            )
            expect_vector = np.matmul(
                np.array(vector_wavefunction).T.conjugate(),
                np.matmul(random_herm, np.array(vector_wavefunction)),
            )

            self.assertAlmostEqual(expect_qiskit.real, expect_vector.real)
            self.assertAlmostEqual(expect_qiskit.imag, expect_vector.imag)
            self.assertAlmostEqual(qiskit_expectation, vector_expectation)

    def test_qaoa_circuit_wavefunction_expectation_equivalence_3(self):
        """Due to the difference in the constructions of the statevector simulators,
        there is a global phase difference between the results obtained from
        qiskit's statevector simulator and OpenQAOA's vectorised simulator. In order to
        show the equivalence between the wavefunctions produced, the expectation
        of a random operator is computed.

        Nonuniform mixer weights.
        """

        nqubits = 3
        p = 1
        weights = [1, 2, 3]
        gammas = [[1 / 8 * np.pi], [1 / 8 * np.pi]]
        betas = [[1], [1 / 8 * np.pi]]

        for i in range(2):
            cost_hamil = Hamiltonian(
                [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
                weights,
                1,
            )

            mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits, coeffs=[1, 2, 3])

            qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
            variate_params = QAOAVariationalStandardParams(
                qaoa_descriptor, betas[i], gammas[i]
            )

            qiskit_backend = QAOAQiskitBackendStatevecSimulator(
                qaoa_descriptor, None, None, True
            )

            qiskit_wavefunction = qiskit_backend.wavefunction(variate_params)
            qiskit_expectation = qiskit_backend.expectation(variate_params)

            vector_backend = QAOAvectorizedBackendSimulator(
                qaoa_descriptor, None, None, True
            )
            vector_wavefunction = vector_backend.wavefunction(variate_params)
            vector_expectation = vector_backend.expectation(variate_params)

            random_operator = (
                np.random.rand(2**nqubits, 2**nqubits)
                + np.random.rand(2**nqubits, 2**nqubits) * 1j
            )
            random_herm = random_operator + random_operator.conj().T

            expect_qiskit = np.matmul(
                np.array(qiskit_wavefunction).T.conjugate(),
                np.matmul(random_herm, np.array(qiskit_wavefunction)),
            )
            expect_vector = np.matmul(
                np.array(vector_wavefunction).T.conjugate(),
                np.matmul(random_herm, np.array(vector_wavefunction)),
            )

            self.assertAlmostEqual(expect_qiskit.real, expect_vector.real)
            self.assertAlmostEqual(expect_qiskit.imag, expect_vector.imag)
            self.assertAlmostEqual(qiskit_expectation, vector_expectation)

    def test_qaoa_circuit_wavefunction_expectation_equivalence_4(self):
        """Due to the difference in the constructions of the statevector simulators,
        there is a global phase difference between the results obtained from
        qiskit's statevector simulator and OpenQAOA's vectorised simulator. In order to
        show the equivalence between the wavefunctions produced, the expectation
        of a random operator is computed.

        Y, YY and XX mixers with nonuniform weights.
        """

        nqubits = 3
        p = 1
        weights = [1, 2, 3]
        gammas = [[1 / 8 * np.pi], [1 / 8 * np.pi]]
        betas = [[1], [1 / 8 * np.pi]]

        for i in range(2):
            cost_hamil = Hamiltonian(
                [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
                weights,
                1,
            )
            mixer_hamil = Hamiltonian(
                [
                    PauliOp("Y", (0,)),
                    PauliOp("YY", (0, 1)),
                    PauliOp("XX", (1, 2)),
                    PauliOp("XZ", (1, 2)),
                ],
                [1, 2, 3, 4],
                1,
            )

            qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
            variate_params = QAOAVariationalStandardParams(
                qaoa_descriptor, betas[i], gammas[i]
            )

            qiskit_backend = QAOAQiskitBackendStatevecSimulator(
                qaoa_descriptor, None, None, True
            )

            qiskit_wavefunction = qiskit_backend.wavefunction(variate_params)
            qiskit_expectation = qiskit_backend.expectation(variate_params)

            vector_backend = QAOAvectorizedBackendSimulator(
                qaoa_descriptor, None, None, True
            )
            vector_wavefunction = vector_backend.wavefunction(variate_params)
            vector_expectation = vector_backend.expectation(variate_params)

            random_operator = (
                np.random.rand(2**nqubits, 2**nqubits)
                + np.random.rand(2**nqubits, 2**nqubits) * 1j
            )
            random_herm = random_operator + random_operator.conj().T

            expect_qiskit = np.matmul(
                np.array(qiskit_wavefunction).T.conjugate(),
                np.matmul(random_herm, np.array(qiskit_wavefunction)),
            )
            expect_vector = np.matmul(
                np.array(vector_wavefunction).T.conjugate(),
                np.matmul(random_herm, np.array(vector_wavefunction)),
            )

            self.assertAlmostEqual(expect_qiskit.real, expect_vector.real)
            self.assertAlmostEqual(expect_qiskit.imag, expect_vector.imag)
            self.assertAlmostEqual(qiskit_expectation, vector_expectation)

    def test_cost_call(self):
        """
        testing the __call__ method of the base class.
        Only for vectorized and Qiskit Local Statevector Backends.
        """

        n_qubits = 8
        register = range(n_qubits)
        p = 1

        betas = [np.pi / 8]
        gammas = [np.pi / 4]
        cost_hamiltonian = ring_of_disagrees(register)
        mixer_hamiltonian = X_mixer_hamiltonian(n_qubits)
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = QAOAVariationalStandardParams(
            qaoa_descriptor, betas, gammas
        )

        backend_qiskit_statevec = QAOAQiskitBackendStatevecSimulator(
            qaoa_descriptor, prepend_state=None, append_state=None, init_hadamard=True
        )

        exp_qiskit_statevec = backend_qiskit_statevec.expectation(
            (variational_params_std)
        )

        assert np.isclose(exp_qiskit_statevec, -6)

    def test_get_wavefunction(self):
        n_qubits = 3
        terms = [[0, 1], [0, 2], [0]]
        weights = [1, 1, -0.5]
        p = 1

        betas_singles = [np.pi, 0, 0]
        betas_pairs = []
        gammas_singles = [np.pi]
        gammas_pairs = [[1 / 2 * np.pi] * 2]

        cost_hamiltonian = Hamiltonian.classical_hamiltonian(
            terms=terms, coeffs=weights, constant=0
        )
        mixer_hamiltonian = X_mixer_hamiltonian(n_qubits)
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = QAOAVariationalExtendedParams(
            qaoa_descriptor,
            betas_singles=betas_singles,
            betas_pairs=betas_pairs,
            gammas_singles=gammas_singles,
            gammas_pairs=gammas_pairs,
        )

        backend_qiskit_statevec = QAOAQiskitBackendStatevecSimulator(
            qaoa_descriptor, prepend_state=None, append_state=None, init_hadamard=True
        )

        wf_qiskit_statevec = backend_qiskit_statevec.wavefunction(
            (variational_params_std)
        )
        expected_wf = 1j * np.array([-1, 1, 1, -1, 1, -1, -1, 1]) / (2 * np.sqrt(2))

        try:
            assert np.allclose(wf_qiskit_statevec, expected_wf)
        except AssertionError:
            assert np.allclose(
                np.real(np.conjugate(wf_qiskit_statevec) * wf_qiskit_statevec),
                np.conjugate(expected_wf) * expected_wf,
            )

    def test_exact_solution(self):
        """
        NOTE:Since the implementation of exact solution is backend agnostic
            Checking it once should be okay.

        Nevertheless, for the sake of completeness it will be tested for all backend
        instances.

        """

        n_qubits = 8
        register = range(n_qubits)
        p = 1

        correct_energy = -8
        correct_config = [0, 1, 0, 1, 0, 1, 0, 1]

        # The tests pass regardless of the value of betas and gammas is this correct?
        betas = [np.pi / 8]
        gammas = [np.pi / 4]

        cost_hamiltonian = ring_of_disagrees(register)
        mixer_hamiltonian = X_mixer_hamiltonian(n_qubits)
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = QAOAVariationalStandardParams(
            qaoa_descriptor, betas, gammas
        )

        backend_qiskit_statevec = QAOAQiskitBackendStatevecSimulator(
            qaoa_descriptor, prepend_state=None, append_state=None, init_hadamard=True
        )

        # exact solution is defined as the property of the cost function
        energy_qiskit, config_qiskit = backend_qiskit_statevec.exact_solution

        assert np.isclose(energy_qiskit, correct_energy)

        config_qiskit = [config.tolist() for config in config_qiskit]

        assert correct_config in config_qiskit

    def test_expectation_w_uncertainty(self):
        """
        Test the standard deviation equality. Expectation w uncertainty.
        """

        n_qubits = 8
        register = range(n_qubits)
        p = 1

        betas = [np.pi / 8]
        gammas = [np.pi / 4]
        cost_hamiltonian = ring_of_disagrees(register)
        mixer_hamiltonian = X_mixer_hamiltonian(n_qubits)
        qaoa_descriptor = QAOADescriptor(cost_hamiltonian, mixer_hamiltonian, p)
        variational_params_std = QAOAVariationalStandardParams(
            qaoa_descriptor, betas, gammas
        )

        backend_qiskit_statevec = QAOAQiskitBackendStatevecSimulator(
            qaoa_descriptor, prepend_state=None, append_state=None, init_hadamard=True
        )

        (
            exp_qiskit_statevec,
            exp_unc_qiskit_statevec,
        ) = backend_qiskit_statevec.expectation_w_uncertainty(variational_params_std)

        assert np.isclose(exp_qiskit_statevec, -6)

        terms = [(register[i], register[(i + 1) % n_qubits]) for i in range(n_qubits)]
        weights = [0.5] * len(terms)

        # Check standard deviation
        # Get the matrix form of the Hamiltonian (note we just keep the diagonal
        # part) and square it
        ham_matrix = np.zeros((2 ** len(register)))
        for i, term in enumerate(terms):
            out = np.real(weights[i])
            for qubit in register:
                if qubit in term:
                    out = np.kron([1, -1], out)
                else:
                    out = np.kron([1, 1], out)
            ham_matrix += out

        ham_matrix_sq = np.square(ham_matrix)

        # Get the wavefunction
        wf = backend_qiskit_statevec.wavefunction(variational_params_std)

        # Get the probabilities
        probs = np.real(np.conjugate(wf) * wf)

        # Standard deviation
        exp_2 = np.dot(probs, ham_matrix)
        std_dev2 = np.sqrt(np.dot(probs, ham_matrix_sq) - exp_2**2)

        assert np.isclose(exp_unc_qiskit_statevec, std_dev2)

    def test_expectation_w_randomizing_variables(self):
        """
        Run ntrials sets of randomized input parameters and compares the
        expectation value output between the qiskit statevector simulator and
        vectorized simulator.
        """

        ntrials = 100

        nqubits = 3
        p = [np.random.randint(1, 4) for i in range(ntrials)]
        weights = [
            [np.random.rand(), np.random.rand(), np.random.rand(), np.random.rand()]
            for i in range(ntrials)
        ]
        init_hadamards = [np.random.choice([True, False]) for i in range(ntrials)]
        constants = [np.random.rand() for i in range(ntrials)]

        for i in range(ntrials):
            gammas = [np.random.rand() * np.pi for i in range(p[i])]
            betas = [np.random.rand() * np.pi for i in range(p[i])]

            cost_hamil = Hamiltonian(
                [
                    PauliOp("Z", (0,)),
                    PauliOp("ZZ", (0, 1)),
                    PauliOp("ZZ", (1, 2)),
                    PauliOp("ZZ", (0, 2)),
                ],
                weights[i],
                constants[i],
            )
            mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
            qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p[i])
            variate_params = QAOAVariationalStandardParams(
                qaoa_descriptor, betas, gammas
            )

            qiskit_backend = QAOAQiskitBackendStatevecSimulator(
                qaoa_descriptor, None, None, init_hadamards[i]
            )

            qiskit_expectation = qiskit_backend.expectation(variate_params)

            vector_backend = QAOAvectorizedBackendSimulator(
                qaoa_descriptor, None, None, init_hadamards[i]
            )
            vector_expectation = vector_backend.expectation(variate_params)

            self.assertAlmostEqual(qiskit_expectation, vector_expectation)

    def test_shot_based_simulator(self):
        """
        Test get_counts in shot-based qiskit simulator.
        """

        nqubits = 3
        p = 1
        weights = [1, 1, 1]
        gammas = [1 / 8 * np.pi]
        betas = [1 / 8 * np.pi]
        shots = 10000

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
        variate_params = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)

        qiskit_shot_backend = QAOAQiskitBackendShotBasedSimulator(
            qaoa_descriptor, shots, None, None, True, 1.0
        )

        shot_result = qiskit_shot_backend.get_counts(variate_params)

        self.assertEqual(type(shot_result), dict)

    def test_cvar_alpha_expectation(self):
        """
        Test computing the expectation value by changing the alpha of the cvar.
        """

        nqubits = 3
        p = 1
        weights = [1, -1, 1]
        gammas = [1 / 8 * np.pi]
        betas = [1 / 8 * np.pi]
        shots = 10000

        cost_hamil = Hamiltonian(
            [PauliOp("ZZ", (0, 1)), PauliOp("ZZ", (1, 2)), PauliOp("ZZ", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
        variate_params = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)

        qiskit_shot_backend = QAOAQiskitBackendShotBasedSimulator(
            qaoa_descriptor, shots, None, None, True, 1.0, seed_simulator=1234
        )

        expectation_value_100 = qiskit_shot_backend.expectation(variate_params)
        self.assertEqual(type(float(expectation_value_100)), float)

        # cvar_alpha = 0.5, 0.75
        qiskit_shot_backend.cvar_alpha = 0.5
        expectation_value_05 = qiskit_shot_backend.expectation(variate_params)
        qiskit_shot_backend.cvar_alpha = 0.75
        expectation_value_075 = qiskit_shot_backend.expectation(variate_params)
        self.assertNotEqual(expectation_value_05, expectation_value_075)
        self.assertEqual(type(float(expectation_value_05)), float)
        self.assertEqual(type(float(expectation_value_075)), float)

    def test_standard_decomposition_branch_in_circuit_construction(self):
        """
        XY Pauli is not an implemented low level gate. Produces NotImplementedError
        as the standard decomposition for the XY PauliGate doesnt exist.
        """

        nqubits = 3
        p = 1
        weights = [1, 1, 1]
        gammas = [1 / 8 * np.pi]
        betas = [1 / 8 * np.pi]
        shots = 10000

        cost_hamil = Hamiltonian(
            [PauliOp("XY", (0, 1)), PauliOp("XY", (1, 2)), PauliOp("XY", (0, 2))],
            weights,
            1,
        )
        mixer_hamil = X_mixer_hamiltonian(n_qubits=nqubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=p)
        variate_params = QAOAVariationalStandardParams(qaoa_descriptor, betas, gammas)

        self.assertRaises(
            NotImplementedError,
            QAOAQiskitBackendShotBasedSimulator,
            qaoa_descriptor,
            shots,
            None,
            None,
            True,
            1.0,
        )

        self.assertRaises(
            NotImplementedError,
            QAOAQiskitBackendStatevecSimulator,
            qaoa_descriptor,
            None,
            None,
            True,
        )


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np

from braket.circuits import gates as braketgates
from braket.circuits import Circuit
from braket.circuits.free_parameter import FreeParameter

from openqaoa.qaoa_components.ansatz_constructor.gates import (
    RY,
    RX,
    RZ,
    CZ,
    CX,
    RXX,
    RYY,
    RZZ,
    RZX,
    CPHASE,
    RiSWAP,
)
from openqaoa.qaoa_components.ansatz_constructor.rotationangle import RotationAngle
from openqaoa_braket.backends.gates_braket import BraketGateApplicator


class TestingGate(unittest.TestCase):
    
    def setUp(self):
        
        self.braket_gate_applicator = BraketGateApplicator()
    
    def test_braket_gates_1q(self):
        
        # Braket Gate Applicator
        gate_applicator = self.braket_gate_applicator

        # One Qubit Gate Tests
        rotation_angle_obj = RotationAngle(lambda x: x, [], FreeParameter("test_angle"))
        
        empty_circuit = Circuit()
        llgate = RY(gate_applicator, 0, rotation_angle_obj)
        output_circuit = llgate.apply_gate(empty_circuit)
        output_circuit = output_circuit.make_bound_circuit({"test_angle": np.pi})

        test_circuit = Circuit()
        test_circuit += braketgates.Ry.ry(0, np.pi)

        self.assertEqual(test_circuit, output_circuit)

        empty_circuit = Circuit()
        llgate = RX(gate_applicator, 0, rotation_angle_obj)
        output_circuit = llgate.apply_gate(empty_circuit)
        output_circuit = output_circuit.make_bound_circuit({"test_angle": np.pi})

        test_circuit = Circuit()
        test_circuit += braketgates.Rx.rx(0, np.pi)

        self.assertEqual(test_circuit, output_circuit)

        empty_circuit = Circuit()
        llgate = RZ(gate_applicator, 0, rotation_angle_obj)
        output_circuit = llgate.apply_gate(empty_circuit)
        output_circuit = output_circuit.make_bound_circuit({"test_angle": np.pi})

        test_circuit = Circuit()
        test_circuit += braketgates.Rz.rz(0, np.pi)

        self.assertEqual(test_circuit, output_circuit)

    def test_braket_gates_2q(self):
        
        # Braket Gate Applicator
        gate_applicator = self.braket_gate_applicator

        # Two Qubit Gate Tests
        empty_circuit = Circuit()
        llgate = CZ(gate_applicator, 0, 1)
        output_circuit = llgate.apply_gate(empty_circuit)

        test_circuit = Circuit()
        test_circuit += braketgates.CZ.cz(0, 1)

        self.assertEqual(test_circuit, output_circuit)

        empty_circuit = Circuit()
        llgate = CX(gate_applicator, 0, 1)
        output_circuit = llgate.apply_gate(empty_circuit)

        test_circuit = Circuit()
        test_circuit += braketgates.CNot.cnot(0, 1)

        self.assertEqual(test_circuit, output_circuit)

    def test_braket_gates_2q_w_gates(self):
        
        # Braket Gate Applicator
        gate_applicator = self.braket_gate_applicator

        # Two Qubit Gate with Angles Tests
        rotation_angle_obj = RotationAngle(lambda x: x, [], FreeParameter("test_angle"))

        empty_circuit = Circuit()
        llgate = RXX(gate_applicator, 0, 1, rotation_angle_obj)
        output_circuit = llgate.apply_gate(empty_circuit)
        output_circuit = output_circuit.make_bound_circuit({"test_angle": np.pi})

        test_circuit = Circuit()
        test_circuit += braketgates.XX.xx(0, 1, np.pi)

        self.assertEqual(test_circuit, output_circuit)

        empty_circuit = Circuit()
        llgate = RYY(gate_applicator, 0, 1, rotation_angle_obj)
        output_circuit = llgate.apply_gate(empty_circuit)
        output_circuit = output_circuit.make_bound_circuit({"test_angle": np.pi})

        test_circuit = Circuit()
        test_circuit += braketgates.YY.yy(0, 1, np.pi)

        self.assertEqual(test_circuit, output_circuit)

        empty_circuit = Circuit()
        llgate = RZZ(gate_applicator, 0, 1, rotation_angle_obj)
        output_circuit = llgate.apply_gate(empty_circuit)
        output_circuit = output_circuit.make_bound_circuit({"test_angle": np.pi})

        test_circuit = Circuit()
        test_circuit += braketgates.ZZ.zz(0, 1, np.pi)

        self.assertEqual(test_circuit, output_circuit)

        empty_circuit = Circuit()
        llgate = CPHASE(gate_applicator, 0, 1, rotation_angle_obj)
        output_circuit = llgate.apply_gate(empty_circuit)
        output_circuit = output_circuit.make_bound_circuit({"test_angle": np.pi})

        test_circuit = Circuit()
        test_circuit += braketgates.CPhaseShift.cphaseshift(0, 1, np.pi)

        self.assertEqual(test_circuit, output_circuit)

        empty_circuit = Circuit()
        llgate = RiSWAP(gate_applicator, 0, 1, rotation_angle_obj)
        output_circuit = llgate.apply_gate(empty_circuit)
        output_circuit = output_circuit.make_bound_circuit({"test_angle": np.pi})

        test_circuit = Circuit()
        test_circuit += braketgates.XY.xy(0, 1, np.pi)

        self.assertEqual(test_circuit, output_circuit)


if __name__ == "__main__":
    unittest.main()



# Copyright 2018-2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTICE #
# Lines modified by Entropica Labs will bear the comment # changed from the original PennyLane code

"""Contains a function that computes the fourier series of
a quantum expectation value."""
from functools import wraps
from inspect import signature
import warnings

import numpy as np
from autoray import numpy as anp
from openqaoa.optimizers import (
    pennylane as qml,
)  # changed from the original PennyLane code


def _reconstruct_equ(fun, num_frequency, x0=None, f0=None, interface=None):
    r"""Reconstruct a univariate Fourier series with consecutive integer
    frequencies, using trigonometric interpolation and equidistant shifts.

    This technique is based on
    `Dirichlet kernels <https://en.wikipedia.org/wiki/Dirichlet_kernel>`_, see
    `Vidal and Theis (2018) <https://arxiv.org/abs/1812.06323>`_ or
    `Wierichs et al. (2022) <https://doi.org/10.22331/q-2022-03-30-677>`_.

    Args:
        fun (callable): Univariate finite Fourier series to reconstruct.
            It must have signature ``float -> float`` .
        num_frequency (int): Number of integer frequencies in ``fun``.
            All integer frequencies below ``num_frequency`` are assumed
            to be present in ``fun`` as well; if they are not, the output
            is correct put the reconstruction could have been performed
            with fewer evaluations of ``fun`` .
        x0 (float): Center to which to shift the reconstruction.
            The points at which ``fun`` is evaluated are *not* affected
            by ``x0`` .
        f0 (float): Value of ``fun`` at zero; Providing ``f0`` saves one
            evaluation of ``fun``.
        interface (str): Which auto-differentiation framework to use as
            interface. This determines in which interface the output
            reconstructed function is intended to be used.

    Returns:
        callable: Reconstructed Fourier series with ``num_frequency`` frequencies.
        This function is a purely classical function. Furthermore, it is fully
        differentiable.
    """
    if not abs(int(num_frequency)) == num_frequency:
        raise ValueError(
            f"num_frequency must be a non-negative integer, got {num_frequency}"
        )

    a = (num_frequency + 0.5) / np.pi
    b = 0.5 / np.pi

    shifts_pos = qml.math.arange(1, num_frequency + 1) / a
    shifts_neg = -shifts_pos[::-1]
    shifts = qml.math.concatenate([shifts_neg, [0.0], shifts_pos])
    shifts = anp.asarray(shifts, like=interface)
    f0 = fun(0.0) if f0 is None else f0
    evals = (
        list(map(fun, shifts[:num_frequency]))
        + [f0]
        + list(map(fun, shifts[num_frequency + 1 :]))
    )
    evals = anp.asarray(evals, like=interface)

    x0 = anp.asarray(np.float64(0.0), like=interface) if x0 is None else x0

    def _reconstruction(x):
        """Univariate reconstruction based on equidistant shifts and Dirichlet kernels.
        The derivative at of ``sinc`` are not well-implemented in TensorFlow and Autograd,
        use the Fourier transform reconstruction if this derivative is needed.
        """
        _x = x - x0 - shifts
        return qml.math.tensordot(
            qml.math.sinc(a * _x) / qml.math.sinc(b * _x),
            evals,
            axes=[[0], [0]],
        )

    return _reconstruction


_warn_text_f0_ignored = (
    "The provided value of the function at zero will be ignored due to the "
    "provided shift values. This may lead to additional evaluations of the "
    "function to be reconstructed."
)


def _reconstruct_gen(fun, spectrum, shifts=None, x0=None, f0=None, interface=None):
    r"""Reconstruct a univariate (real-valued) Fourier series with given spectrum.

    Args:
        fun (callable): Univariate finite Fourier series to reconstruct.
            It must have signature ``float -> float`` .
        spectrum (Collection): Frequency spectrum of the Fourier series;
            non-positive frequencies are ignored.
        shifts (Sequence): Shift angles at which to evaluate ``fun`` for the reconstruction.
            Chosen equidistantly within the interval :math:`[0, 2\pi/f_\text{max}]`
            if ``shifts=None`` , where :math:`f_\text{max}` is the biggest
            frequency in ``spectrum``.
        x0 (float): Center to which to shift the reconstruction.
            The points at which ``fun`` is evaluated are *not* affected
            by ``x0`` .
        f0 (float): Value of ``fun`` at zero; If :math:`0` is among the ``shifts``
            and ``f0`` is provided, one evaluation of ``fun`` is saved.
        interface (str): Which auto-differentiation framework to use as
            interface. This determines in which interface the output
            reconstructed function is intended to be used.

    Returns:
        callable: Reconstructed Fourier series with :math:`R` frequencies in ``spectrum`` .
        This function is a purely classical function. Furthermore, it is fully differentiable.
    """
    # pylint: disable=unused-argument, too-many-arguments

    have_f0 = f0 is not None
    have_shifts = shifts is not None

    spectrum = anp.asarray(spectrum, like=interface)
    spectrum = spectrum[spectrum > 0]
    f_max = qml.math.max(spectrum)

    # If no shifts are provided, choose equidistant ones
    if not have_shifts:
        R = qml.math.shape(spectrum)[0]
        shifts = qml.math.arange(-R, R + 1) * 2 * np.pi / (f_max * (2 * R + 1)) * R
        zero_idx = R
        need_f0 = True
    elif have_f0:
        zero_idx = qml.math.where(
            qml.math.isclose(shifts, qml.math.zeros_like(shifts[0]))
        )
        zero_idx = (
            zero_idx[0][0] if (len(zero_idx) > 0 and len(zero_idx[0]) > 0) else None
        )
        need_f0 = zero_idx is not None

    # Take care of shifts close to zero if f0 was provided
    if have_f0 and need_f0:
        # Only one shift may be zero at a time
        shifts = qml.math.concatenate(
            [shifts[zero_idx : zero_idx + 1], shifts[:zero_idx], shifts[zero_idx + 1 :]]
        )
        shifts = anp.asarray(shifts, like=interface)
        evals = anp.asarray([f0] + list(map(fun, shifts[1:])), like=interface)
    else:
        shifts = anp.asarray(shifts, like=interface)
        if have_f0 and not need_f0:
            warnings.warn(_warn_text_f0_ignored)
        evals = anp.asarray(list(map(fun, shifts)), like=interface)

    L = len(shifts)
    # Construct the coefficient matrix case by case
    C1 = qml.math.ones((L, 1))
    C2 = qml.math.cos(qml.math.tensordot(shifts, spectrum, axes=0))
    C3 = qml.math.sin(qml.math.tensordot(shifts, spectrum, axes=0))
    C = qml.math.hstack([C1, C2, C3])

    # Solve the system of linear equations
    cond = qml.math.linalg.cond(C)
    if cond > 1e8:
        warnings.warn(
            f"The condition number of the Fourier transform matrix is very large: {cond}.",
            UserWarning,
        )
    W = qml.math.linalg.solve(C, evals)

    # Extract the Fourier coefficients
    R = (L - 1) // 2
    a0 = W[0]
    a = anp.asarray(W[1 : R + 1], like=interface)
    b = anp.asarray(W[R + 1 :], like=interface)

    x0 = anp.asarray(np.float64(0.0), like=interface) if x0 is None else x0

    # Construct the Fourier series
    def _reconstruction(x):
        """Univariate reconstruction based on arbitrary shifts."""
        x = x - x0
        return (
            a0
            + qml.math.tensordot(qml.math.cos(spectrum * x), a, axes=[[0], [0]])
            + qml.math.tensordot(qml.math.sin(spectrum * x), b, axes=[[0], [0]])
        )

    return _reconstruction


def _parse_ids(ids, info_dict):
    """Parse different formats of ``ids`` into the right dictionary format,
    potentially using the information in ``info_dict`` to complete it.
    """
    if ids is None:
        # Infer all id information from info_dict
        return {
            outer_key: inner_dict.keys() for outer_key, inner_dict in info_dict.items()
        }
    if isinstance(ids, str):
        # ids only provides a single argument name but no parameter indices
        return {ids: info_dict[ids].keys()}
    if not isinstance(ids, dict):
        # ids only provides argument names but no parameter indices
        return {_id: info_dict[_id].keys() for _id in ids}

    return ids


def _parse_shifts(shifts, R, arg_name, par_idx, atol, need_f0):
    """Processes shifts for a single reconstruction and determines
    wheter the function at the reconstruction point, ``f0`` will be
    needed.
    """
    # pylint: disable=too-many-arguments
    _shifts = shifts.get(arg_name)
    if _shifts is not None:
        _shifts = _shifts.get(par_idx)
    if _shifts is not None:
        # Check whether the _shifts have the correct size
        if len(_shifts) != 2 * R + 1:
            raise ValueError(
                f"The number of provided shifts ({len(_shifts)}) does not fit to the "
                f"number of frequencies (2R+1={2*R+1}) for parameter {par_idx} in "
                f"argument {arg_name}."
            )
        if any(
            qml.math.isclose(_shifts, qml.math.zeros_like(_shifts), rtol=0, atol=atol)
        ):
            # If 0 is among the shifts, f0 is needed
            return _shifts, True
        # If 0 is not among the shifts, f0 is not needed
        return _shifts, (False or need_f0)
    # If no shifts are given, f0 is needed always
    return _shifts, True


def _prepare_jobs(ids, nums_frequency, spectra, shifts, atol):
    r"""For inputs to reconstruct, determine how the given information yields
    function reconstruction tasks and collect them into a dictionary ``jobs``.
    Also determine whether the function at zero is needed.

    Args:
        ids (dict or Sequence or str): Indices for the QNode parameters with respect to which
            the QNode should be reconstructed as a univariate function, per QNode argument.
            Each key of the dict, entry of the list, or the single ``str`` has to be the name
            of an argument of ``qnode`` .
            If a ``dict`` , the values of ``ids`` have to contain the parameter indices
            for the respective array-valued QNode argument represented by the key.
            These indices always are tuples, i.e. ``()`` for scalar and ``(i,)`` for
            one-dimensional arguments.
            If a ``list`` , the parameter indices are inferred from ``nums_frequency`` if
            given or ``spectra`` else.
            If ``None``, all keys present in ``nums_frequency`` / ``spectra`` are considered.
        nums_frequency (dict[dict]): Numbers of integer frequencies -- and biggest
            frequency -- per QNode parameter. The keys have to be argument names of ``qnode``
            and the inner dictionaries have to be mappings from parameter indices to the
            respective integer number of frequencies. If the QNode frequencies are not contiguous
            integers, the argument ``spectra`` should be used to save evaluations of ``qnode`` .
            Takes precedence over ``spectra`` and leads to usage of equidistant shifts.
        spectra (dict[dict]): Frequency spectra per QNode parameter.
            The keys have to be argument names of ``qnode`` and the inner dictionaries have to
            be mappings from parameter indices to the respective frequency spectrum for that
            parameter. Ignored if ``nums_frequency!=None``.
        shifts (dict[dict]): Shift angles for the reconstruction per QNode parameter.
            The keys have to be argument names of ``qnode`` and the inner dictionaries have to
            be mappings from parameter indices to the respective shift angles to be used for that
            parameter. For :math:`R` non-zero frequencies, there must be :math:`2R+1` shifts
            given. Ignored if ``nums_frequency!=None``.
        atol (float): Absolute tolerance used to analyze shifts lying close to 0.

    Returns:
        dict[dict]: Indices for the QNode parameters with respect to which the QNode
            will be reconstructed. Cast to the dictionary structure explained above.
            If the input ``ids`` was a dictionary, it is returned unmodified.
        callable: The reconstruction method to use, one out of two internal methods.
        dict[dict[dict]]: Keyword arguments for the reconstruction method specifying
            how to carry out the reconstruction. The outer-most keys are QNode argument
            names, the middle keys are parameter indices like the inner keys of
            ``nums_frequency`` or ``spectra`` and the inner-most dictionary contains the
            keyword arguments, i.e. the keys are keyword argument names for the
            reconstruction method
        bool: Whether any of the reconstruction jobs will require the evaluation
            of the function at the position of reconstruction itself.
    """
    if nums_frequency is None:
        if spectra is None:
            raise ValueError("Either nums_frequency or spectra must be given.")

        ids = _parse_ids(ids, spectra)

        if shifts is None:
            shifts = {}

        need_f0 = False
        recon_fn = _reconstruct_gen

        jobs = {}

        # If no shifts are provided, compute them
        for arg_name, inner_dict in ids.items():
            _jobs = {}

            for par_idx in inner_dict:
                # Determine spectrum and number of frequencies, discounting for 0
                _spectrum = spectra[arg_name][par_idx]
                R = len(_spectrum) - 1
                _shifts, need_f0 = _parse_shifts(
                    shifts, R, arg_name, par_idx, atol, need_f0
                )

                # Store job
                if R > 0:
                    _jobs[par_idx] = {"shifts": _shifts, "spectrum": _spectrum}
                else:
                    # R=0 belongs to a constant function
                    _jobs[par_idx] = None

            jobs[arg_name] = _jobs

    else:
        jobs = {}
        need_f0 = True

        ids = _parse_ids(ids, nums_frequency)

        recon_fn = _reconstruct_equ

        for arg_name, inner_dict in ids.items():
            _jobs = {}

            for par_idx in inner_dict:
                _num_frequency = nums_frequency[arg_name][par_idx]
                _jobs[par_idx] = (
                    {"num_frequency": _num_frequency} if _num_frequency > 0 else None
                )

            jobs[arg_name] = _jobs

    return ids, recon_fn, jobs, need_f0


def reconstruct(qnode, ids=None, nums_frequency=None, spectra=None, shifts=None):
    r"""Reconstruct an expectation value QNode along a single parameter direction.
    This means we restrict the QNode to vary only one parameter, a univariate restriction.
    For common quantum gates, such restrictions are finite Fourier series with known
    frequency spectra. Thus they may be reconstructed using Dirichlet kernels or
    a non-uniform Fourier transform.

    Args:
        qnode (pennylane.QNode): Quantum node to be reconstructed, representing a
            circuit that outputs an expectation value.
        ids (dict or Sequence or str): Indices for the QNode parameters with respect to which
            the QNode should be reconstructed as a univariate function, per QNode argument.
            Each key of the dict, entry of the list, or the single ``str`` has to be the name
            of an argument of ``qnode`` .
            If a ``dict`` , the values of ``ids`` have to contain the parameter indices
            for the respective array-valued QNode argument represented by the key.
            These indices always are tuples, i.e., ``()`` for scalar and ``(i,)`` for
            one-dimensional arguments.
            If a ``list`` , the parameter indices are inferred from ``nums_frequency`` if
            given or ``spectra`` else.
            If ``None``, all keys present in ``nums_frequency`` / ``spectra`` are considered.
        nums_frequency (dict[dict]): Numbers of integer frequencies -- and biggest
            frequency -- per QNode parameter. The keys have to be argument names of ``qnode``
            and the inner dictionaries have to be mappings from parameter indices to the
            respective integer number of frequencies. If the QNode frequencies are not contiguous
            integers, the argument ``spectra`` should be used to save evaluations of ``qnode`` .
            Takes precedence over ``spectra`` and leads to usage of equidistant shifts.
        spectra (dict[dict]): Frequency spectra per QNode parameter.
            The keys have to be argument names of ``qnode`` and the inner dictionaries have to
            be mappings from parameter indices to the respective frequency spectrum for that
            parameter. Ignored if ``nums_frequency!=None``.
        shifts (dict[dict]): Shift angles for the reconstruction per QNode parameter.
            The keys have to be argument names of ``qnode`` and the inner dictionaries have to
            be mappings from parameter indices to the respective shift angles to be used for that
            parameter. For :math:`R` non-zero frequencies, there must be :math:`2R+1` shifts
            given. Ignored if ``nums_frequency!=None``.

    Returns:
        function: Function which accepts the same arguments as the QNode and one additional
        keyword argument ``f0`` to provide the QNode value at the given arguments.
        When called, this function will return a dictionary of dictionaries,
        formatted like ``nums_frequency`` or ``spectra`` ,
        that contains the univariate reconstructions per QNode parameter.

    For each provided ``id`` in ``ids``, the QNode is restricted to varying the single QNode
    parameter corresponding to the ``id`` . This univariate function is then reconstructed
    via a Fourier transform or Dirichlet kernels, depending on the provided input.
    Either the frequency ``spectra`` of the QNode with respect to its input parameters or
    the numbers of frequencies, ``nums_frequency`` , per parameter must be provided.

    For quantum-circuit specific details, we refer the reader to
    `Vidal and Theis (2018) <https://arxiv.org/abs/1812.06323>`__ ,
    `Vidal and Theis (2020) <https://www.frontiersin.org/articles/10.3389/fphy.2020.00297/full>`__ ,
    `Schuld, Sweke and Meyer (2021) <https://journals.aps.org/pra/abstract/10.1103/PhysRevA.103.032430>`__ ,
    and
    `Wierichs, Izaac, Wang and Lin (2022) <https://doi.org/10.22331/q-2022-03-30-677>`__ .
    An introduction to the concept of quantum circuits as Fourier series can also be found in
    the
    `Quantum models as Fourier series <https://pennylane.ai/qml/demos/tutorial_expressivity_fourier_series.html>`__
    and
    `General parameter-shift rules <https://pennylane.ai/qml/demos/tutorial_general_parshift.html>`__
    demos as well as the
    :mod:`qml.fourier <pennylane.fourier>` module docstring.

    **Example**

    Consider the following QNode:

    .. code-block:: python

        dev = qml.device("default.qubit", wires=2)

        @qml.qnode(dev)
        def circuit(x, Y):
            qml.RX(x, wires=0)
            qml.RY(Y[0], wires=0)
            qml.RY(Y[1], wires=1)
            qml.CNOT(wires=[0, 1])
            qml.RY(5*  Y[1], wires=1)
            return qml.expval(qml.PauliZ(0) @ qml.PauliZ(1))

        x = 0.4
        Y = np.array([1.9, -0.5])
        f = 2.3

        circuit_value = circuit(x, Y)

    It has three variational parameters ``x`` (a scalar) and two entries of ``Y``
    (an array-like).
    A reconstruction job could then be with respect to the two entries of ``Y``,
    which enter the circuit with one and six integer frequencies, respectively
    (see the additional examples below for details on how to obtain the frequency
    spectrum if it is not known):

    >>> nums_frequency = {"Y": {(0,): 1, (1,): 6}}
    >>> with qml.Tracker(circuit.device) as tracker:
    ...     rec = qml.fourier.reconstruct(circuit, {"Y": [(0,), (1,)]}, nums_frequency)(x, Y)
    >>> rec.keys()
    dict_keys(['Y'])
    >>> print(*rec["Y"].items(), sep="\n")
    ((0,), <function _reconstruct_equ.<locals>._reconstruction at 0x7fbd685aee50>)
    ((1,), <function _reconstruct_equ.<locals>._reconstruction at 0x7fbd6866eee0>)
    >>> recon_Y0 = rec["Y"][(0,)]
    >>> recon_Y1 = rec["Y"][(1,)]
    >>> np.isclose(recon_Y0(Y[0]), circuit_value)
    True
    >>> np.isclose(recon_Y1(Y[1]+1.3), circuit(x, Y+np.eye(2)[1]*1.3))
    True

    We successfully reconstructed the dependence on the two entries of ``Y`` ,
    keeping ``x`` and the respective other entry in ``Y`` at their initial values.
    Let us also see how many executions of the device were used to obtain the
    reconstructions:

    >>> tracker.totals
    {'executions': 15}

    The example above used that we already knew the frequency spectra of the
    QNode of interest. However, this is in general not the case and we may need
    to compute the spectrum first. This can be done with
    :func:`.fourier.qnode_spectrum` :

    >>> spectra = qml.fourier.qnode_spectrum(circuit)(x, Y)
    >>> spectra.keys()
    dict_keys(['x', 'Y'])
    >>> spectra["x"]
    {(): [-1.0, 0.0, 1.0]}
    >>> print(*spectra["Y"].items(), sep="\n")
    ((0,), [-1.0, 0.0, 1.0])
    ((1,), [-6.0, -5.0, -4.0, -1.0, 0.0, 1.0, 4.0, 5.0, 6.0])

    For more detailed explanations, usage details and additional examples, see
    the usage details section below.

    .. details::
        :title: Usage Details

        **Input formatting**

        As described briefly above, the essential inputs to ``reconstruct`` that provide information
        about the QNode are given as dictionaries of dictionaries, where the outer keys reference
        the argument names of ``qnode`` and the inner keys reference the parameter indices within
        each array-valued QNode argument. These parameter indices always are tuples, so that
        for scalar-valued QNode parameters, the parameter index is ``()`` by convention and the
        ``i`` -th parameter of a one-dimensional array can be accessed via ``(i,)`` .
        For example, providing ``nums_frequency``

        - for a scalar argument: ``nums_frequency = {"x": {(): 4}}``
        - for a one-dimensional argument: ``nums_frequency = {"Y": {(0,): 2, (1,): 9, (4,): 1}}``
        - for a three-dimensional argument: ``nums_frequency = {"Z": {(0, 2, 5): 2, (1, 1, 4): 1}}``

        This applies to ``nums_frequency`` , ``spectra`` , and ``shifts`` .

        Note that the information provided in ``nums_frequency`` / ``spectra`` is essential for
        the correctness of the reconstruction.

        On the other hand, the input format for ``ids`` is flexible and allows a collection of
        parameter indices for each QNode argument name (as a ``dict`` ), a collection of argument
        names (as a ``list``, ``set``, ``tuple`` or similar), or a single argument name
        (as a ``str`` ) to be defined. For ``ids=None`` , all argument names contained in
        ``nums_frequency`` -- or ``spectra`` if ``nums_frequency`` is not used -- are considered.
        For inputs that do not specify parameter indices per QNode argument name (all formats but
        ``dict`` ), these parameter indices are inferred from ``nums_frequency`` / ``spectra`` .

        **Reconstruction cost**

        The reconstruction cost -- in terms of calls to ``qnode`` -- depend on the number of
        frequencies given via ``nums_frequency`` or ``spectra`` . A univariate reconstruction
        for :math:`R` frequencies takes :math:`2R+1` evaluations. If multiple univariate
        reconstructions are performed at the same point with various numbers of frequencies
        :math:`R_k` , the cost are :math:`1+2\sum_k R_k` if the shift :math:`0` is used in all
        of them. This is in particular the case if ``nums_frequency`` or ``spectra`` with
        ``shifts=None`` is used.

        If the number of frequencies is too large or the given frequency spectrum contains
        more than the spectrum of ``qnode`` , the reconstruction is performed suboptimally
        but remains correct.
        For integer-valued spectra with gaps, the equidistant reconstruction is thus suboptimal
        and the non-equidistant version method be used (also see the examples below).

        **Numerical stability**

        In general, the reconstruction with equidistant shifts for equidistant frequencies
        (used if ``nums_frequency`` is provided) is more stable numerically than the more
        general Fourier reconstruction (used if ``nums_frequency=None`` ).
        If the system of equations to be solved in the Fourier transform is
        ill-conditioned, a warning is raised as the output might become unstable.
        Examples for this are shift values or frequencies that lie very close to each other.

        **Differentiability**

        The returned scalar functions are differentiable in all interfaces with respect
        to their scalar input variable. They expect these inputs to be in the same
        interface as the one used by the QNode. More advanced differentiability, for example
        of the reconstructions with respect to QNode properties, is not supported
        reliably yet.

        .. warning::

            When using ``TensorFlow`` or ``Autograd`` *and* ``nums_frequency`` ,
            the reconstructed functions are not differentiable at the point of
            reconstruction. One workaround for this is to use ``spectra`` as
            input instead and to thereby use the Fourier transform instead of
            Dirichlet kernels. Alternatively, the original QNode evaluation can
            be used.

        **More examples**

        Consider the QNode from the example above, now with an additional, tunable frequency
        ``f`` for the Pauli-X rotation that is controlled by ``x`` :

        .. code-block:: python

            @qml.qnode(dev)
            def circuit(x, Y, f=1.0):
                qml.RX(f * x, wires=0)
                qml.RY(Y[0], wires=0)
                qml.RY(Y[1], wires=1)
                qml.CNOT(wires=[0, 1])
                qml.RY(5*  Y[1], wires=1)
                return qml.expval(qml.PauliZ(0) @ qml.PauliZ(1))

            f = 2.3

            circuit_value = circuit(x, Y)


        We repeat the reconstruction job for the dependence on ``Y[1]`` .
        Note that even though information about ``Y[0]`` is contained in ``nums_frequency`` ,
        ``ids`` determines which reconstructions are performed.

        >>> with qml.Tracker(circuit.device) as tracker:
        ...     rec = qml.fourier.reconstruct(circuit, {"Y": [(1,)]}, nums_frequency)(x, Y)
        >>> tracker.totals
        {'executions': 13}

        As expected, we required :math:`2R+1=2\cdot 6+1=13` circuit executions. However, not
        all frequencies below :math:`f_\text{max}=6` are present in the circuit, so that
        a reconstruction using knowledge of the full frequency spectrum will be cheaper:

        >>> spectra = {"Y": {(1,): [0., 1., 4., 5., 6.]}}
        >>> with tracker:
        ...     rec = qml.fourier.reconstruct(circuit, {"Y": [(1,)]}, None, spectra)(x, Y)
        >>> tracker.totals
        {'executions': 9}

        We again obtain the full univariate dependence on ``Y[1]`` but with considerably
        fewer executions on the quantum device.
        Once we obtained the classical function that describes the dependence, no
        additional circuit evaluations are performed:

        >>> with tracker:
        ...     for Y1 in np.arange(-np.pi, np.pi, 20):
        ...         rec["Y"][(1,)](-2.1)
        >>> tracker.totals
        {}

        If we want to reconstruct the dependence of ``circuit`` on ``x`` , we cannot use
        ``nums_frequency`` if ``f`` is not an integer. One could rescale ``x`` to obtain
        the frequency :math:`1` again, or directly use ``spectra`` . We will combine the
        latter with another reconstruction with respect to ``Y[0]`` :

        >>> spectra = {"x": {(): [0., f]}, "Y": {(0,): [0., 1.]}}
        >>> with tracker:
        ...     rec = qml.fourier.reconstruct(circuit, None, None, spectra)(x, Y, f=f)
        >>> tracker.totals
        {'executions': 5}
        >>> recon_x = rec["x"][()]
        >>> np.isclose(recon_x(x+0.5), circuit(x+0.5, Y, f=f)
        True

        Note that by convention, the parameter index for a scalar variable is ``()`` and
        that the frequency :math:`0` always needs to be included in the spectra.
        Furthermore, we here skipped the input ``ids`` so that the reconstruction
        was performed for all keys in ``spectra`` .
        The reconstruction with a single non-zero frequency
        costs three evaluations of ``circuit`` for each, ``x`` and ``Y[0]`` . Performing
        both reconstructions at the same position allowed us to save one of the
        evaluations and reduce the number of calls to :math:`5`.

    """
    # pylint: disable=cell-var-from-loop, unused-argument

    atol = 1e-8
    ids, recon_fn, jobs, need_f0 = _prepare_jobs(
        ids, nums_frequency, spectra, shifts, atol
    )
    sign_fn = qnode.func if isinstance(qnode, qml.QNode) else qnode
    arg_names = list(signature(sign_fn).parameters.keys())
    arg_idx_from_names = {arg_name: i for i, arg_name in enumerate(arg_names)}

    @wraps(qnode)
    def wrapper(*args, f0=None, **kwargs):
        if f0 is None and need_f0:
            f0 = qnode(*args, **kwargs)

        interface = qml.math.get_interface(args[0])

        def constant_fn(x):
            """Univariate reconstruction of a constant Fourier series."""
            return f0

        # Carry out the reconstruction jobs
        reconstructions = {}
        for arg_name, inner_dict in jobs.items():
            _reconstructions = {}
            arg_idx = arg_idx_from_names[arg_name]

            for par_idx, job in inner_dict.items():
                if job is None:
                    _reconstructions[par_idx] = constant_fn
                else:
                    if len(qml.math.shape(args[arg_idx])) == 0:
                        shift_vec = qml.math.ones_like(args[arg_idx])
                        x0 = args[arg_idx]
                    else:
                        shift_vec = qml.math.zeros_like(args[arg_idx])
                        shift_vec = qml.math.scatter_element_add(
                            shift_vec, par_idx, 1.0
                        )
                        x0 = args[arg_idx][par_idx]

                    def _univariate_fn(x):
                        new_arg = args[arg_idx] + shift_vec * x
                        new_args = args[:arg_idx] + (new_arg,) + args[arg_idx + 1 :]
                        return qnode(*new_args, **kwargs)

                    _reconstructions[par_idx] = recon_fn(
                        _univariate_fn, **job, x0=x0, f0=f0, interface=interface
                    )

            reconstructions[arg_name] = _reconstructions

        return reconstructions

    return wrapper



import unittest
import pytest

from openqaoa.backends.qaoa_backend import (
    DEVICE_ACCESS_OBJECT_MAPPER
)
from openqaoa.qaoa_components import (
    Hamiltonian,
    create_qaoa_variational_params,
    QAOADescriptor,
)
from openqaoa.utilities import X_mixer_hamiltonian
from openqaoa_qiskit.backends import DeviceQiskit

def get_params():
    cost_hamil = Hamiltonian.classical_hamiltonian([[0, 1]], [1], constant=0)
    mixer_hamil = X_mixer_hamiltonian(2)

    qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)
    variational_params_std = create_qaoa_variational_params(
        qaoa_descriptor, "standard", "ramp"
    )

    return qaoa_descriptor, variational_params_std


class TestingBackendQPUs(unittest.TestCase):
    """
    These tests check methods of the QPU backends.
    """

    @pytest.mark.qpu
    def setUp(self):

        self.HUB = "ibm-q"
        self.GROUP = "open"
        self.PROJECT = "main"

    @pytest.mark.qpu
    def test_get_counts_and_expectation_n_shots(self):
        """
        TODO: test needs to be updated as DEVICE_ACCESS_OBJECT_MAPPER is now dynamically filled based on whether a module exists.
        
        Check that the .get_counts, .expectation and .expecation_w_uncertainty methods admit n_shots as an argument for the backends of all QPUs.
        """

        list_device_attributes = [
            {
                "QPU": "Qiskit",
                "device_name": "ibmq_qasm_simulator",
                "hub": self.HUB,
                "group": self.GROUP,
                "project": self.PROJECT,
            }
        ]

        assert DeviceQiskit in DEVICE_ACCESS_OBJECT_MAPPER.keys()
        
        device = DeviceQiskit
        backend = DEVICE_ACCESS_OBJECT_MAPPER[DeviceQiskit]
        device_attributes = list_device_attributes[0]

        qaoa_descriptor, variational_params_std = get_params()

        QPU_name = device_attributes.pop("QPU")
        print("Testing {} backend.".format(QPU_name))

        # if QPU_name in ["AWS", 'Qiskit']:
        #     print(f"Skipping test for {QPU_name} backend.")
        #     continue

        try:
            print(device, device_attributes)
            device = device(**device_attributes)
            backend = backend(
                qaoa_descriptor=qaoa_descriptor,
                device=device,
                cvar_alpha=1,
                n_shots=100,
                prepend_state=None,
                append_state=None,
                init_hadamard=True,
            )

            # Check that the .get_counts, .expectation and .expectation_w_variance methods admit n_shots as an argument
            assert (
                sum(
                    backend.get_counts(
                        params=variational_params_std, n_shots=58
                    ).values()
                )
                == 58
            ), "`n_shots` is not being respected when calling .get_counts(n_shots=58).".format(
                QPU_name
            )
            backend.expectation(params=variational_params_std, n_shots=58)
            backend.expectation_w_uncertainty(
                params=variational_params_std, n_shots=58
            )

        except Exception as e:

            raise e from type(e)(f"Error raised for `{QPU_name}`: " + str(e))

        print("Test passed for {} backend.".format(QPU_name))


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
from openqaoa.problems import NumberPartition


def terms_list_equality(terms_list1, terms_list2):
    """
    Check the terms equality between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = True if (term1 == term2 or term1 == term2[::-1]) else False

    return bool


def terms_list_isclose(terms_list1, terms_list2):
    """
    Check if the distance between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = (
                True
                if np.isclose(term1, term2) or np.isclose(term1, term2[::-1])
                else False
            )

    return bool

class TestNumberPartition(unittest.TestCase):
    """Tests for NumberPartition class"""
    def test_number_partitioning_terms_weights_constant(self):
        """Test that Number Partitioning creates the correct terms, weights, constant"""
        list_numbers = [1, 2, 3]
        expected_terms = [[0, 1], [0, 2], [1, 2]]
        expected_weights = [4, 6, 12]
        expected_constant = 14

        np_problem = NumberPartition(list_numbers)
        qubo_problem = np_problem.qubo

        self.assertTrue(terms_list_equality(qubo_problem.terms, expected_terms))
        self.assertEqual(qubo_problem.weights, expected_weights)
        self.assertEqual(qubo_problem.constant, expected_constant)

    def test_number_partitioning_random_problem(self):
        """Test randomly generated NumberPartition problem"""
        # regenerate the same numbers randomly
        rng = np.random.default_rng(1234)
        random_numbers_list = list(map(int, rng.integers(1, 10, size=5)))
        manual_np_prob = NumberPartition(random_numbers_list).qubo

        np_prob_random = NumberPartition.random_instance(n_numbers=5, seed=1234).qubo

        self.assertTrue(terms_list_equality(np_prob_random.terms, manual_np_prob.terms))
        self.assertEqual(np_prob_random.weights, manual_np_prob.weights)
        self.assertEqual(np_prob_random.constant, manual_np_prob.constant)

    def test_num_part_type_checking(self):
        """
        Checks if the type-checking returns the right error.
        """

        # numbers type-check
        numbers_list = [(1, 2), {"test": 1, "test1": 2}, np.array([1, 2])]

        for each_number in numbers_list:
            with self.assertRaises(TypeError) as e:
                NumberPartition(numbers=each_number)
            self.assertEqual(
                "The input parameter, numbers, has to be a list", str(e.exception)
            )

        numbers_list = [[0.1, 1], [np.array(1), np.array(2)]]

        for each_number in numbers_list:
            with self.assertRaises(TypeError) as e:
                NumberPartition(numbers=each_number)
            self.assertEqual(
                "The elements in numbers list must be of type int.", str(e.exception)
            )


if __name__ == "__main__":
    unittest.main()



# unit testing for circuit routing functionality in OQ
import unittest
from typing import List, Optional
import pytest

from openqaoa import QAOA
from openqaoa.backends import create_device
from openqaoa.problems import NumberPartition, QUBO, Knapsack
from openqaoa.backends.devices_core import DeviceBase
from openqaoa.qaoa_components.ansatz_constructor.gatemap import SWAPGateMap


class ExpectedRouting:
    def __init__(
        self,
        qubo,
        device_name,
        device_location,
        qpu_credentials,
        problem_to_solve,
        initial_mapping,
        gate_indices_list,
        swap_mask,
        initial_physical_to_logical_mapping,
        final_logical_qubit_order,
    ):
        self.qubo = qubo
        self.device = create_device(
            name=device_name, location=device_location, **qpu_credentials
        )

        self.device_name = device_name
        self.device_location = device_location
        self.problem_to_solve = problem_to_solve
        self.initial_mapping = initial_mapping

        self.gate_indices_list = gate_indices_list
        self.swap_mask = swap_mask
        self.initial_physical_to_logical_mapping = initial_physical_to_logical_mapping
        self.final_logical_qubit_order = final_logical_qubit_order

    def values_input(self):
        return (
            self.device_name,
            self.device_location,
            self.problem_to_solve,
            self.initial_mapping,
        )

    def values_return(self):
        return (
            self.gate_indices_list,
            self.swap_mask,
            self.initial_physical_to_logical_mapping,
            self.final_logical_qubit_order,
        )


class TestingQubitRouting(unittest.TestCase):
    @pytest.mark.qpu
    def setUp(self):

        # case qubits device > qubits problem (IBM NAIROBI)
        self.IBM_NAIROBI_KNAPSACK = ExpectedRouting(
            qubo=Knapsack.random_instance(n_items=3, seed=20).qubo,
            device_location="ibmq",
            device_name="ibm_nairobi",
            qpu_credentials={
                "hub": "ibm-q",
                "group": "open",
                "project": "main",
                "as_emulator": True,
            },
            problem_to_solve=[
                (0, 1),
                (2, 3),
                (2, 4),
                (3, 4),
                (0, 2),
                (0, 3),
                (0, 4),
                (1, 2),
                (1, 3),
                (1, 4),
            ],
            initial_mapping=None,
            gate_indices_list=[
                [2, 4],
                [1, 2],
                [3, 5],
                [0, 5],
                [4, 5],
                [4, 5],
                [2, 4],
                [0, 5],
                [2, 4],
                [1, 2],
                [4, 5],
                [0, 5],
                [1, 2],
                [2, 4],
                [2, 4],
                [0, 5],
                [4, 5],
            ],
            swap_mask=[
                False,
                False,
                True,
                False,
                False,
                True,
                False,
                False,
                True,
                False,
                True,
                False,
                True,
                False,
                True,
                True,
                False,
            ],
            initial_physical_to_logical_mapping={6: 0, 2: 1, 1: 2, 4: 3, 3: 4, 5: 5},
            final_logical_qubit_order=[5, 4, 0, 1, 2, 3],
        )

        # case qubits problem == 2 (IBM OSLO)
        self.IBM_OSLO_QUBO2 = ExpectedRouting(
            qubo=QUBO.from_dict(
                {
                    "terms": [[0, 1], [1]],
                    "weights": [9.800730090617392, 26.220558065741773],
                    "n": 2,
                }
            ),
            device_location="ibmq",
            device_name="ibm_oslo",
            qpu_credentials={
                "hub": "ibm-q",
                "group": "open",
                "project": "main",
                "as_emulator": True,
            },
            problem_to_solve=[(0, 1)],
            initial_mapping=None,
            gate_indices_list=[[0, 2], [1, 2]],
            swap_mask=[True, False],
            initial_physical_to_logical_mapping={2: 0, 3: 1, 1: 2},
            final_logical_qubit_order=[2, 1, 0],
        )

        # case qubits device == qubits problem (IBM OSLO)
        self.IBM_OSLO_NPARTITION = ExpectedRouting(
            qubo=NumberPartition.random_instance(n_numbers=7, seed=2).qubo,
            device_location="ibmq",
            device_name="ibm_oslo",
            qpu_credentials={
                "hub": "ibm-q",
                "group": "open",
                "project": "main",
                "as_emulator": True,
            },
            problem_to_solve=[
                (0, 1),
                (0, 2),
                (0, 3),
                (0, 4),
                (0, 5),
                (0, 6),
                (1, 2),
                (1, 3),
                (1, 4),
                (1, 5),
                (1, 6),
                (2, 3),
                (2, 4),
                (2, 5),
                (2, 6),
                (3, 4),
                (3, 5),
                (3, 6),
                (4, 5),
                (4, 6),
                (5, 6),
            ],
            initial_mapping=None,
            gate_indices_list=[
                [0, 5],
                [1, 5],
                [3, 4],
                [4, 5],
                [2, 3],
                [3, 6],
                [4, 5],
                [0, 5],
                [1, 5],
                [3, 4],
                [3, 4],
                [2, 3],
                [3, 6],
                [4, 5],
                [0, 5],
                [1, 5],
                [2, 3],
                [3, 6],
                [3, 4],
                [3, 4],
                [3, 6],
                [0, 5],
                [1, 5],
                [4, 5],
                [4, 5],
                [1, 5],
                [3, 6],
                [3, 4],
                [3, 4],
                [1, 5],
                [4, 5],
            ],
            swap_mask=[
                False,
                False,
                False,
                False,
                False,
                False,
                True,
                False,
                False,
                False,
                True,
                False,
                False,
                True,
                False,
                False,
                True,
                False,
                False,
                True,
                False,
                True,
                False,
                False,
                True,
                False,
                True,
                False,
                True,
                True,
                False,
            ],
            initial_physical_to_logical_mapping={
                0: 0,
                2: 1,
                4: 2,
                5: 3,
                3: 4,
                1: 5,
                6: 6,
            },
            final_logical_qubit_order=[3, 5, 1, 0, 6, 2, 4],
        )

        # create a list of all the cases
        self.list_of_cases = []
        for value in self.__dict__.values():
            if isinstance(value, ExpectedRouting):
                self.list_of_cases.append(value)

    def __routing_function_mock(
        self,
        device: DeviceBase,
        problem_to_solve: List[List[int]],
        initial_mapping: Optional[List[int]] = None,
    ):
        """
        function that imitates the routing function for testing purposes.
        """

        for case in self.list_of_cases:
            if case.values_input() == (
                device.device_name,
                device.device_location,
                problem_to_solve,
                initial_mapping,
            ):
                return case.values_return()

        raise ValueError(
            """The input values are not in the list of expected values, 
                            check the expected cases and the input values. 
                            The input values are: device_name: {}, device_location: {}, problem_to_solve: {}, 
                            initial_mapping: {}""".format(
                device.device_name,
                device.device_location,
                problem_to_solve,
                initial_mapping,
            )
        )

    def __compare_results(self, expected: ExpectedRouting, p: int):
        """
        function that runs qaoa with the routing function and the problem to solve and
        compares the expected and actual results of the routing function.

        :param expected: ExpectedRouting object that contains the input and the expected results of the routing function
        :param p: number of layers of the qaoa circuit
        """
        print(f"\t Testing p={p}")

        device = expected.device
        qubo = expected.qubo

        qaoa = QAOA()
        qaoa.set_device(device)
        qaoa.set_circuit_properties(
            p=p, param_type="standard", init_type="rand", mixer_hamiltonian="x"
        )
        qaoa.set_backend_properties(prepend_state=None, append_state=None)
        qaoa.set_classical_optimizer(
            method="nelder-mead",
            maxiter=1,
            cost_progress=True,
            parameter_log=True,
            optimization_progress=True,
        )
        qaoa.compile(qubo, routing_function=self.__routing_function_mock)
        qaoa.optimize()

        backend, result = qaoa.backend, qaoa.result

        # do the checks

        assert backend.n_qubits == len(
            expected.final_logical_qubit_order
        ), """Number of qubits in the circuit is not equal to the number of qubits given by routing"""

        assert (
            backend.problem_qubits == qubo.n
        ), f"""Number of nodes in problem is not equal to backend.problem_qubits, 
        is '{backend.problem_qubits }' but should be '{qubo.n}'"""

        assert (
            len(list(result.optimized["measurement_outcomes"].keys())[0]) == qubo.n
        ), "The number of qubits in the optimized circuit is not equal to the number of qubits in the problem."

        # check that swap gates are applied in the correct position
        swap_mask_new = []
        for gate in backend.abstract_circuit:
            if not gate.gate_label.n_qubits == 1:
                swap_mask_new.append(isinstance(gate, SWAPGateMap))

        # create the expected swap mask (for p==2 the second swap mask is reversed)
        expected_swap_mask = []
        for i in range(p):
            expected_swap_mask += expected.swap_mask[:: (-1) ** (i % 2)]

        assert (
            swap_mask_new == expected_swap_mask
        ), "Swap gates are not in the correct position"

        # check that the correct qubits are used in the gates
        gate_indices_list_new = []
        for gate in backend.abstract_circuit:
            if not gate.gate_label.n_qubits == 1:
                gate_indices_list_new.append([gate.qubit_1, gate.qubit_2])

        # create the expected swap mask (for p==2 the second swap mask is reversed)
        expected_gate_indices_list = []
        for i in range(p):
            expected_gate_indices_list += expected.gate_indices_list[:: (-1) ** (i % 2)]

        assert (
            gate_indices_list_new == expected_gate_indices_list
        ), "The qubits used in the gates are not correct"

    @pytest.mark.qpu
    def test_qubit_routing(self):

        for i, case in enumerate(self.list_of_cases):
            print("Test case {} out of {}:".format(i + 1, len(self.list_of_cases)))
            self.__compare_results(case, p=i % 4 + 1)
            print("Test passed for case: {}".format(case.values_input()))


if __name__ == "__main__":
    unittest.main()



"""
#   Measure_Lib support will added in a future version. Particularly essential
#   when the Hamiltionian have non-commuting terms.

# from entropica_qaoa.vqe.measurelib import (append_measure_register,
#                                            commuting_decomposition,
#                                            sampling_expectation,
#                                            kron_eigs,
#                                            base_change_fun,
#                                            wavefunction_expectation)
"""
from abc import ABC, abstractmethod, abstractproperty
from typing import Union, List, Dict, Optional, Any, Tuple
from copy import deepcopy
import numpy as np

from .basedevice import DeviceBase
from ..qaoa_components import (
    GateMap,
    QAOADescriptor,
)
from ..qaoa_components.variational_parameters.variational_baseparams import (
    QAOAVariationalBaseParams,
)
from ..utilities import qaoa_probabilities, round_value
from .cost_function import cost_function


class QuantumCircuitBase:
    """
    Phantom class to indicate Quantum Circuits constructed using
    several acceptable services. For instance, IBMQ, PyQuil
    """

    pass


class VQABaseBackend(ABC):
    """
    This is the Abstract Base Class over which other classes will be built.
    Since, this is an Abstract Base class, in order to prevent its initialisation
    the class methods -- ``__init__`` and ``__cal__`` will be decorated as
    `abstractmethods`.

    The Child classes MUST implement and override these abstract methods in their
    implementation specific to their needs.

    NOTE:
        In addition one can also implement other methods which are not
        necessitated by the ``VQABaseBackend`` Base Class


    Parameters
    ----------
    prepend_state: `Union[QuantumCircuitBase, List[complex], np.ndarray]`
        The initial state to start the quantum circuit in the backend.
    append_state: `Union[QuantumCircuitBase, np.ndarray]`
        The final state to append to the quantum circuit in the backend.
    """

    @abstractmethod
    def __init__(
        self,
        prepend_state: Optional[Union[QuantumCircuitBase, List[complex], np.ndarray]],
        append_state: Optional[Union[QuantumCircuitBase, np.ndarray]],
    ):
        """The constructor. See class docstring"""
        self.prepend_state = prepend_state
        self.append_state = append_state

    @abstractmethod
    def expectation(self, params: Any) -> float:
        """
        Call the execute function on the circuit to compute the
        expectation value of the Quantum Circuit w.r.t cost operator
        """
        pass

    @abstractmethod
    def expectation_w_uncertainty(self, params: Any) -> Tuple[float, float]:
        """
        Call the execute function on the circuit to compute the
        expectation value of the Quantum Circuit w.r.t cost operator
        along with its uncertainty
        """
        pass

    @abstractproperty
    def exact_solution(self):
        """
        Use linear algebra to compute the exact solution of the problem
        Hamiltonian classically.
        """
        pass


class QAOABaseBackend(VQABaseBackend):
    """
    This class inherits from the VQABaseBackend and needs to be backend
    agnostic to QAOA implementations on different devices and their
    respective SDKs.

    Parameters
    ----------
    qaoa_descriptor: `QAOADescriptor`
        This object handles the information to design the QAOA circuit ansatz
    prepend_state: `Union[QuantumCircuitBase, List[complex]]`
        Warm Starting the QAOA problem with some initial state other than the regular
        $|+ \\rangle ^{otimes n}$
    append_state: `Union[QuantumCircuitBase, List[complex]]`
        Appending a user-defined circuit/state to the end of the QAOA routine
    init_hadamard: `bool`
        Initialises the QAOA circuit with a hadamard when ``True``
    cvar_alpha: `float`
    """

    def __init__(
        self,
        qaoa_descriptor: QAOADescriptor,
        prepend_state: Optional[Union[QuantumCircuitBase, List[complex], np.ndarray]],
        append_state: Optional[Union[QuantumCircuitBase, np.ndarray]],
        init_hadamard: bool,
        cvar_alpha: float,
    ):
        super().__init__(prepend_state, append_state)

        self.qaoa_descriptor = qaoa_descriptor
        self.cost_hamiltonian = qaoa_descriptor.cost_hamiltonian
        self.n_qubits = self.qaoa_descriptor.n_qubits
        self.init_hadamard = init_hadamard
        self.cvar_alpha = cvar_alpha
        self.problem_qubits = self.qaoa_descriptor.cost_hamiltonian.n_qubits

        self.abstract_circuit = deepcopy(self.qaoa_descriptor.abstract_circuit)

        # pass the generated mappings if the circuit is routed
        if self.qaoa_descriptor.routed == True:
            self.initial_qubit_mapping = self.qaoa_descriptor.initial_mapping

            if self.qaoa_descriptor.p % 2 != 0:
                self.final_mapping = self.qaoa_descriptor.final_mapping
            else:
                # if even, the initial mapping [0,...,n_qubits-1] is taken as the final mapping
                self.final_mapping = list(
                    range(len(self.qaoa_descriptor.final_mapping))
                )
        else:
            self.initial_qubit_mapping = None
            self.final_mapping = None

    def assign_angles(self, params: QAOAVariationalBaseParams) -> None:
        """
        Assigns the angle values of the variational parameters to the circuit gates
        specified as a list of gates in the ``abstract_circuit``.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The variational parameters(angles) to be assigned to the circuit gates
        """
        # if circuit is non-parameterised, then assign the angle values to the circuit
        abstract_circuit = self.abstract_circuit

        for each_gate in abstract_circuit:
            gate_label_layer = each_gate.gate_label.layer
            gate_label_seq = each_gate.gate_label.sequence
            if each_gate.gate_label.n_qubits == 2:
                if each_gate.gate_label.type.value == "MIXER":
                    angle = params.mixer_2q_angles[gate_label_layer, gate_label_seq]
                elif each_gate.gate_label.type.value == "COST":
                    angle = params.cost_2q_angles[gate_label_layer, gate_label_seq]
            elif each_gate.gate_label.n_qubits == 1:
                if each_gate.gate_label.type.value == "MIXER":
                    angle = params.mixer_1q_angles[gate_label_layer, gate_label_seq]
                elif each_gate.gate_label.type.value == "COST":
                    angle = params.cost_1q_angles[gate_label_layer, gate_label_seq]
            each_gate.angle_value = angle

        self.abstract_circuit = abstract_circuit

    def obtain_angles_for_pauli_list(
        self, input_gate_list: List[GateMap], params: QAOAVariationalBaseParams
    ) -> List[float]:
        """
        This method uses the pauli gate list information to obtain the pauli angles
        from the VariationalBaseParams object. The floats in the list are in the order
        of the input GateMaps list.

        Parameters
        ----------
        input_gate_list: `List[GateMap]`
            The GateMap list including rotation gates
        params: `QAOAVariationalBaseParams`
            The variational parameters(angles) to be assigned to the circuit gates

        Returns
        -------
        angles_list: `List[float]`
            The list of angles in the order of gates in the `GateMap` list
        """
        angle_list = []

        for each_gate in input_gate_list:
            gate_label_layer = each_gate.gate_label.layer
            gate_label_seq = each_gate.gate_label.sequence

            if each_gate.gate_label.n_qubits == 2:
                if each_gate.gate_label.type.value == "MIXER":
                    angle_list.append(
                        params.mixer_2q_angles[gate_label_layer, gate_label_seq]
                    )
                elif each_gate.gate_label.type.value == "COST":
                    angle_list.append(
                        params.cost_2q_angles[gate_label_layer, gate_label_seq]
                    )
            elif each_gate.gate_label.n_qubits == 1:
                if each_gate.gate_label.type.value == "MIXER":
                    angle_list.append(
                        params.mixer_1q_angles[gate_label_layer, gate_label_seq]
                    )
                elif each_gate.gate_label.type.value == "COST":
                    angle_list.append(
                        params.cost_1q_angles[gate_label_layer, gate_label_seq]
                    )

        return angle_list

    @abstractmethod
    def qaoa_circuit(self, params: QAOAVariationalBaseParams) -> QuantumCircuitBase:
        """
        Construct the QAOA circuit and append the parameter values to obtain the final
        circuit ready for execution on the device.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters as a 1D array (derived from an object of one of
            the parameter classes, containing hyperparameters and variable parameters).

        Returns
        -------
        quantum_circuit: `QuantumCircuitBase`
            A Quantum Circuit object of type created by the respective
            backend service
        """
        pass

    @abstractmethod
    def get_counts(self, params: QAOAVariationalBaseParams, n_shots=None) -> dict:
        """
        This method will be implemented in the child classes according to the type
        of backend used.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters - an object of one of the parameter classes, containing
            variable parameters.
        n_shots: `int`
            The number of shots to be used for the measurement. If None, the backend default.
        """
        pass

    @round_value
    def expectation(self, params: QAOAVariationalBaseParams, n_shots=None) -> float:
        """
        Compute the expectation value w.r.t the Cost Hamiltonian

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters - an object of one of the parameter classes, containing
            variable parameters.
        n_shots: `int`
            The number of shots to be used for the measurement. If None, the backend default.

        Returns
        -------
        float:
            Expectation value of cost operator wrt to quantum state produced by QAOA circuit
        """
        counts = self.get_counts(params, n_shots)
        cost = cost_function(
            counts, self.qaoa_descriptor.cost_hamiltonian, self.cvar_alpha
        )
        return cost

    @round_value
    def expectation_w_uncertainty(
        self, params: QAOAVariationalBaseParams, n_shots=None
    ) -> Tuple[float, float]:
        """
        Compute the expectation value w.r.t the Cost Hamiltonian and its uncertainty

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters - an object of one of the parameter classes, containing
            variable parameters.
        n_shots: `int`
            The number of shots to be used for the measurement. If None, the backend default.

        Returns
        -------
        Tuple[float]:
            expectation value and its uncertainty of cost operator wrt
            to quantum state produced by QAOA circuit.
        """
        counts = self.get_counts(params, n_shots)
        cost = cost_function(
            counts, self.qaoa_descriptor.cost_hamiltonian, self.cvar_alpha
        )
        cost_sq = cost_function(
            counts,
            self.qaoa_descriptor.cost_hamiltonian.hamiltonian_squared,
            self.cvar_alpha,
        )

        uncertainty = np.sqrt(cost_sq - cost**2)

        return (cost, uncertainty)

    @abstractmethod
    def reset_circuit(self):
        """
        Reset the circuit attribute
        """
        pass

    @property
    def exact_solution(self):
        """
        Computes exactly the minimum energy of the cost function and its
        corresponding configuration of variables using standard numpy module.

        Returns
        -------
        (energy, config): `Tuple[float, list]`
            - The minimum eigenvalue of the cost Hamiltonian,
            - The minimum energy eigenvector as a binary array
              configuration: qubit-0 as the first element in the sequence
        """
        register = self.qaoa_descriptor.qureg
        terms = self.cost_hamiltonian.terms
        coeffs = self.cost_hamiltonian.coeffs
        constant_energy = self.cost_hamiltonian.constant

        diag = np.zeros((2 ** len(register)))
        for i, term in enumerate(terms):
            out = np.real(coeffs[i])
            for qubit in register:
                if qubit in term.qubit_indices:
                    out = np.kron([1, -1], out)
                else:
                    out = np.kron([1, 1], out)
            diag += out

        # add the constant energy contribution
        diag += constant_energy

        # index = np.argmin(diag)
        energy = np.min(diag)
        indices = []
        for idx in range(len(diag)):
            if diag[idx] == energy:
                indices.append(idx)

        config_strings = [
            np.binary_repr(index, len(register))[::-1] for index in indices
        ]
        configs = [
            np.array([int(x) for x in config_str]) for config_str in config_strings
        ]

        return energy, configs

    def bitstring_energy(self, bitstring: Union[List[int], str]) -> float:
        """
        Computes the energy of a given bitstring with respect to the cost Hamiltonian.

        Parameters
        ----------
        bitstring : `Union[List[int],str]`
            A list of integers 0 and 1 of length `n_qubits` representing a configuration.

        Returns
        -------
        float:
            The energy of a given bitstring with respect to the cost Hamiltonian.
        """
        energy = 0
        string_rev = bitstring
        terms = self.cost_hamiltonian.terms
        coeffs = self.cost_hamiltonian.coeffs
        constant_energy = self.cost_hamiltonian.constant

        for i, term in enumerate(terms):
            variables_product = np.prod([(-1) ** string_rev[k] for k in term])
            energy += coeffs[i] * variables_product
        energy += constant_energy

        return energy

    @abstractmethod
    def circuit_to_qasm(self):
        """
        Implement a method to construct a QASM string from the current
        state of the QuantumCircuit for the backends
        """
        pass


class QAOABaseBackendStatevector(QAOABaseBackend):
    """
    Base backend class for a statevector simulator backend
    """

    @abstractmethod
    def wavefunction(self, params: QAOAVariationalBaseParams) -> List[complex]:
        """
        Get the wavefunction of the state produced by the QAOA circuit.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters - an object of one of the parameter classes, containing
            the variational parameters (angles).

        Returns
        -------
        List[complex]:
            A list of the wavefunction amplitudes.
        """
        pass

    def sample_from_wavefunction(
        self, params: QAOAVariationalBaseParams, n_samples: int
    ) -> np.ndarray:
        """
        Get the shot-based measurement results from the statevector. The return object is
        a list of shot-results.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters as a 1D array (derived from an object of one of the
            parameter classes, containing hyperparameters and variable parameters).
        n_samples: `int`
            The number of measurement samples required; specified as integer

        Returns
        -------
        np.ndarray:
            A list of measurement outcomes sampled from a statevector
        """
        wf = self.wavefunction(params)
        prob_vec = np.real(np.conjugate(wf) * wf)
        samples = np.random.choice(len(prob_vec), p=prob_vec, size=n_samples)
        samples = [np.binary_repr(num, self.n_qubits)[::-1] for num in samples]

        return samples

    def probability_dict(self, params: QAOAVariationalBaseParams):
        """
        Get the counts style probability dictionary with all basis states
        and their corresponding probabilities. Constructed using the complete
        statevector

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters as a 1D array (derived from an object of one of the
            parameter classes, containing hyperparameters and variable parameters).

        Returns
        -------
        Dict[str, float]:
            A dictionary of all basis states and their corresponding probabilities.
        """

        wf = self.wavefunction(params)
        return qaoa_probabilities(wf)

    def get_counts(self, params: QAOAVariationalBaseParams, n_shots: int) -> Dict:
        """
        Measurement outcome vs frequency information from a circuit execution
        represented as a python dictionary

        Parameters
        ----------
        params: `VariationalBaseParams`
            The QAOA parameters as a 1D array (derived from an object of one of the
            parameter classes, containing hyperparameters and variable parameters).
        n_shots: `int`
            The number of measurement shots required; specified as integer

        Returns
        -------
        Dict[str, float]:
            A dictionary of measurement outcomes vs frequency sampled from a statevector
        """
        samples = self.sample_from_wavefunction(params, n_shots)

        unique_nums, frequency = np.unique(samples, return_counts=True)
        counts = dict(zip(unique_nums, frequency))

        return counts


class QAOABaseBackendShotBased(QAOABaseBackend):
    """
    Implementation of Backend object specific to shot-based simulators and QPUs
    """

    def __init__(
        self,
        qaoa_descriptor: QAOADescriptor,
        n_shots: int,
        prepend_state: Optional[QuantumCircuitBase],
        append_state: Optional[QuantumCircuitBase],
        init_hadamard: bool,
        cvar_alpha: float,
    ):
        super().__init__(
            qaoa_descriptor, prepend_state, append_state, init_hadamard, cvar_alpha
        )
        # assert self.n_qubits >= len(prepend_state.qubits), \
        # "Cannot attach a bigger circuit to the QAOA routine"
        # assert self.n_qubits >= len(append_state.qubits), \
        # "Cannot attach a bigger circuit to the QAOA routine"
        self.n_shots = n_shots

    @abstractmethod
    def get_counts(self, params: QAOAVariationalBaseParams, n_shots=None) -> dict:
        """
        Measurement outcome vs frequency information from a circuit execution
        represented as a python dictionary

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters as a 1D array (derived from an object of one of the
            parameter classes, containing hyperparameters and variable parameters).
        n_shots: `int`
            The number of shots to be used for the measurement. If None, the backend default.

        Returns
        -------
        Dict[str, float]:
            A dictionary of measurement outcomes vs frequency sampled from a statevector
        """
        pass


class QAOABaseBackendCloud:
    """
    QAOA backend that can be accessed over the cloud offered by the
    respective provider through an API based access
    """

    def __init__(self, device: DeviceBase):
        self.device = device
        if getattr(self.device, "provider_connected", None) is None and getattr(self.device, "qpu_connected", None) is None:
            self.device.check_connection()


class QAOABaseBackendParametric:
    """
    Base class to indicate Parametric Circuit Backend
    """

    @abstractmethod
    def parametric_qaoa_circuit(self):
        pass



# Copyright 2018-2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# NOTICE #
# Lines modified by Entropica Labs will bear the comment # changed from the original PennyLane code

"""Differentiable quantum functions"""
# pylint: disable=import-outside-toplevel
import itertools
import functools

from string import ascii_letters as ABC
from autoray import numpy as np
from numpy import float64

from openqaoa.optimizers import (
    pennylane as qml,
)  # changed from the original PennyLane code

from . import single_dispatch  # pylint:disable=unused-import
from .multi_dispatch import diag, dot, scatter_element_add, einsum, get_interface
from .utils import is_abstract, allclose, cast, convert_like, cast_like

ABC_ARRAY = np.array(list(ABC))


def cov_matrix(prob, obs, wires=None, diag_approx=False):
    """Calculate the covariance matrix of a list of commuting observables, given
    the joint probability distribution of the system in the shared eigenbasis.

    .. note::
        This method only works for **commuting observables.**
        If the probability distribution is the result of a quantum circuit,
        the quantum state must be rotated into the shared
        eigenbasis of the list of observables before measurement.

    Args:
        prob (tensor_like): probability distribution
        obs (list[.Observable]): a list of observables for which
            to compute the covariance matrix
        diag_approx (bool): if True, return the diagonal approximation
        wires (.Wires): The wire register of the system. If not provided,
            it is assumed that the wires are labelled with consecutive integers.

    Returns:
        tensor_like: the covariance matrix of size ``(len(obs), len(obs))``

    **Example**

    Consider the following ansatz and observable list:

    >>> obs_list = [qml.PauliX(0) @ qml.PauliZ(1), qml.PauliY(2)]
    >>> ansatz = qml.templates.StronglyEntanglingLayers

    We can construct a QNode to output the probability distribution in the shared eigenbasis of the
    observables:

    .. code-block:: python

        dev = qml.device("default.qubit", wires=3)

        @qml.qnode(dev, interface="autograd")
        def circuit(weights):
            ansatz(weights, wires=[0, 1, 2])
            # rotate into the basis of the observables
            for o in obs_list:
                o.diagonalizing_gates()
            return qml.probs(wires=[0, 1, 2])

    We can now compute the covariance matrix:

    >>> shape = qml.templates.StronglyEntanglingLayers.shape(n_layers=2, n_wires=3)
    >>> weights = np.random.random(shape, requires_grad=True)
    >>> cov = qml.math.cov_matrix(circuit(weights), obs_list)
    >>> cov
    array([[0.98707611, 0.03665537],
         [0.03665537, 0.99998377]])

    Autodifferentiation is fully supported using all interfaces.
    Here we use autograd:

    >>> cost_fn = lambda weights: qml.math.cov_matrix(circuit(weights), obs_list)[0, 1]
    >>> qml.grad(cost_fn)(weights)[0]
    array([[[ 4.94240914e-17, -2.33786398e-01, -1.54193959e-01],
            [-3.05414996e-17,  8.40072236e-04,  5.57884080e-04],
            [ 3.01859411e-17,  8.60411436e-03,  6.15745204e-04]],
           [[ 6.80309533e-04, -1.23162742e-03,  1.08729813e-03],
            [-1.53863193e-01, -1.38700657e-02, -1.36243323e-01],
            [-1.54665054e-01, -1.89018172e-02, -1.56415558e-01]]])
    """
    variances = []

    # diagonal variances
    for i, o in enumerate(obs):
        eigvals = cast(o.eigvals(), dtype=float64)
        w = o.wires.labels if wires is None else wires.indices(o.wires)
        p = marginal_prob(prob, w)

        res = dot(eigvals**2, p) - (dot(eigvals, p)) ** 2
        variances.append(res)

    cov = diag(variances)

    if diag_approx:
        return cov

    for i, j in itertools.combinations(range(len(obs)), r=2):
        o1 = obs[i]
        o2 = obs[j]

        o1wires = o1.wires.labels if wires is None else wires.indices(o1.wires)
        o2wires = o2.wires.labels if wires is None else wires.indices(o2.wires)
        shared_wires = set(o1wires + o2wires)

        l1 = cast(o1.eigvals(), dtype=float64)
        l2 = cast(o2.eigvals(), dtype=float64)
        l12 = cast(np.kron(l1, l2), dtype=float64)

        p1 = marginal_prob(prob, o1wires)
        p2 = marginal_prob(prob, o2wires)
        p12 = marginal_prob(prob, shared_wires)

        res = dot(l12, p12) - dot(l1, p1) * dot(l2, p2)

        cov = scatter_element_add(cov, [i, j], res)
        cov = scatter_element_add(cov, [j, i], res)

    return cov


def marginal_prob(prob, axis):
    """Compute the marginal probability given a joint probability distribution expressed as a tensor.
    Each random variable corresponds to a dimension.

    If the distribution arises from a quantum circuit measured in computational basis, each dimension
    corresponds to a wire. For example, for a 2-qubit quantum circuit `prob[0, 1]` is the probability of measuring the
    first qubit in state 0 and the second in state 1.

    Args:
        prob (tensor_like): 1D tensor of probabilities. This tensor should of size
            ``(2**N,)`` for some integer value ``N``.
        axis (list[int]): the axis for which to calculate the marginal
            probability distribution

    Returns:
        tensor_like: the marginal probabilities, of
        size ``(2**len(axis),)``

    **Example**

    >>> x = tf.Variable([1, 0, 0, 1.], dtype=tf.float64) / np.sqrt(2)
    >>> marginal_prob(x, axis=[0, 1])
    <tf.Tensor: shape=(4,), dtype=float64, numpy=array([0.70710678, 0.        , 0.        , 0.70710678])>
    >>> marginal_prob(x, axis=[0])
    <tf.Tensor: shape=(2,), dtype=float64, numpy=array([0.70710678, 0.70710678])>
    """
    prob = np.flatten(prob)
    num_wires = int(np.log2(len(prob)))

    if num_wires == len(axis):
        return prob

    inactive_wires = tuple(set(range(num_wires)) - set(axis))
    prob = np.reshape(prob, [2] * num_wires)
    prob = np.sum(prob, axis=inactive_wires)
    return np.flatten(prob)


def _density_matrix_from_matrix(density_matrix, indices, check_state=False):
    """Compute the density matrix from a state represented with a density matrix.


    Args:
        density_matrix (tensor_like): 2D density matrix tensor. This tensor should be of size ``(2**N, 2**N)`` for some
            integer number of wires``N``.
        indices (list(int)): List of indices in the considered subsystem.
        check_state (bool): If True, the function will check the state validity (shape and norm).

    Returns:
        tensor_like: Density matrix of size ``(2**len(wires), 2**len(wires))``

    **Example**

    >>> x = np.array([[1, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]])
    >>> _density_matrix_from_matrix(x, indices=[0])
    [[1.+0.j 0.+0.j]
     [0.+0.j 0.+0.j]]

    >>> y = [[0.5, 0, 0.5, 0], [0, 0, 0, 0], [0.5, 0, 0.5, 0], [0, 0, 0, 0]]
    >>> _density_matrix_from_matrix(y, indices=[0])
    [[0.5+0.j 0.5+0.j]
     [0.5+0.j 0.5+0.j]]

    >>> _density_matrix_from_matrix(y, indices=[1])
    [[1.+0.j 0.+0.j]
     [0.+0.j 0.+0.j]]

    >>> z = tf.Variable([[1, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]], dtype=tf.complex128)
    >>> _density_matrix_from_matrix(x, indices=[1])
    tf.Tensor(
    [[1.+0.j 0.+0.j]
     [0.+0.j 0.+0.j]], shape=(2, 2), dtype=complex128)


    """
    shape = density_matrix.shape[0]
    num_indices = int(np.log2(shape))

    if check_state:
        _check_density_matrix(density_matrix)

    consecutive_indices = list(range(0, num_indices))

    # Return the full density matrix if all the wires are given
    if tuple(indices) == tuple(consecutive_indices):
        return density_matrix

    traced_wires = [x for x in consecutive_indices if x not in indices]
    density_matrix = _partial_trace(density_matrix, traced_wires)
    return density_matrix


def _partial_trace(density_matrix, indices):
    """Compute the reduced density matrix by tracing out the provided indices.

    Args:
        density_matrix (tensor_like): 2D density matrix tensor. This tensor should be of size ``(2**N, 2**N)`` for some
            integer number of wires ``N``.
        indices (list(int)): List of indices to be traced.

    Returns:
        tensor_like: (reduced) Density matrix of size ``(2**len(wires), 2**len(wires))``

    **Example**

    >>> x = np.array([[1, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]])
    >>> _partial_trace(x, indices=[0])
    [[1.+0.j 0.+0.j]
     [0.+0.j 0.+0.j]]


    >>> x = tf.Variable([[1, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]], dtype=tf.complex128)
    >>> _partial_trace(x, indices=[1])
    tf.Tensor(
    [[1.+0.j 0.+0.j]
     [0.+0.j 0.+0.j]], shape=(2, 2), dtype=complex128)
    """
    # Autograd does not support same indices sum in backprop
    if get_interface(density_matrix) == "autograd":
        density_matrix = _partial_trace_autograd(density_matrix, indices)
        return density_matrix

    # Dimension and reshape
    shape = density_matrix.shape[0]
    num_indices = int(np.log2(shape))
    rho_dim = 2 * num_indices

    density_matrix = np.reshape(density_matrix, [2] * 2 * num_indices)
    indices = np.sort(indices)

    # For loop over wires
    for i, target_index in enumerate(indices):
        target_index = target_index - i
        state_indices = ABC[: rho_dim - 2 * i]
        state_indices = list(state_indices)

        target_letter = state_indices[target_index]
        state_indices[target_index + num_indices - i] = target_letter
        state_indices = "".join(state_indices)

        einsum_indices = f"{state_indices}"
        density_matrix = einsum(einsum_indices, density_matrix)

    number_wires_sub = num_indices - len(indices)
    reduced_density_matrix = np.reshape(
        density_matrix, (2**number_wires_sub, 2**number_wires_sub)
    )
    return reduced_density_matrix


def _partial_trace_autograd(density_matrix, indices):
    """Compute the reduced density matrix for autograd interface by tracing out the provided indices with the use
    of projectors as same subscripts indices are not supported in autograd backprop.
    """
    # Dimension and reshape
    shape = density_matrix.shape[0]
    num_indices = int(np.log2(shape))
    rho_dim = 2 * num_indices
    density_matrix = np.reshape(density_matrix, [2] * 2 * num_indices)

    kraus = cast(np.eye(2), density_matrix.dtype)

    kraus = np.reshape(kraus, (2, 1, 2))
    kraus_dagger = np.asarray([np.conj(np.transpose(k)) for k in kraus])

    kraus = convert_like(kraus, density_matrix)
    kraus_dagger = convert_like(kraus_dagger, density_matrix)
    # For loop over wires
    for target_wire in indices:
        # Tensor indices of density matrix
        state_indices = ABC[:rho_dim]
        # row indices of the quantum state affected by this operation
        row_wires_list = [target_wire]
        row_indices = "".join(ABC_ARRAY[row_wires_list].tolist())
        # column indices are shifted by the number of wires
        col_wires_list = [w + num_indices for w in row_wires_list]
        col_indices = "".join(ABC_ARRAY[col_wires_list].tolist())
        # indices in einsum must be replaced with new ones
        num_partial_trace_wires = 1
        new_row_indices = ABC[rho_dim : rho_dim + num_partial_trace_wires]
        new_col_indices = ABC[
            rho_dim + num_partial_trace_wires : rho_dim + 2 * num_partial_trace_wires
        ]
        # index for summation over Kraus operators
        kraus_index = ABC[
            rho_dim
            + 2 * num_partial_trace_wires : rho_dim
            + 2 * num_partial_trace_wires
            + 1
        ]
        # new state indices replace row and column indices with new ones
        new_state_indices = functools.reduce(
            lambda old_string, idx_pair: old_string.replace(idx_pair[0], idx_pair[1]),
            zip(col_indices + row_indices, new_col_indices + new_row_indices),
            state_indices,
        )
        # index mapping for einsum, e.g., 'iga,abcdef,idh->gbchef'
        einsum_indices = (
            f"{kraus_index}{new_row_indices}{row_indices}, {state_indices},"
            f"{kraus_index}{col_indices}{new_col_indices}->{new_state_indices}"
        )
        density_matrix = einsum(einsum_indices, kraus, density_matrix, kraus_dagger)

    number_wires_sub = num_indices - len(indices)
    reduced_density_matrix = np.reshape(
        density_matrix, (2**number_wires_sub, 2**number_wires_sub)
    )
    return reduced_density_matrix


def _density_matrix_from_state_vector(state, indices, check_state=False):
    """Compute the density matrix from a state vector.

    Args:
        state (tensor_like): 1D tensor state vector. This tensor should of size ``(2**N,)`` for some integer value ``N``.
        indices (list(int)): List of indices in the considered subsystem.
        check_state (bool): If True, the function will check the state validity (shape and norm).

    Returns:
        tensor_like: Density matrix of size ``(2**len(indices), 2**len(indices))``

    **Example**

    >>> x = np.array([1, 0, 0, 0])
    >>> _density_matrix_from_state_vector(x, indices=[0])
    [[1.+0.j 0.+0.j]
    [0.+0.j 0.+0.j]]

    >>> y = [1, 0, 1, 0] / np.sqrt(2)
    >>> _density_matrix_from_state_vector(y, indices=[0])
    [[0.5+0.j 0.5+0.j]
     [0.5+0.j 0.5+0.j]]

    >>> _density_matrix_from_state_vector(y, indices=[1])
    [[1.+0.j 0.+0.j]
     [0.+0.j 0.+0.j]]

    >>> z = tf.Variable([1, 0, 0, 0], dtype=tf.complex128)
    >>> _density_matrix_from_state_vector(z, indices=[1])
    tf.Tensor(
    [[1.+0.j 0.+0.j]
     [0.+0.j 0.+0.j]], shape=(2, 2), dtype=complex128)

    """
    len_state = np.shape(state)[0]

    # Check the format and norm of the state vector
    if check_state:
        _check_state_vector(state)

    # Get dimension of the quantum system and reshape
    num_indices = int(np.log2(len_state))
    consecutive_wires = list(range(num_indices))
    state = np.reshape(state, [2] * num_indices)

    # Get the system to be traced
    traced_system = [x for x in consecutive_wires if x not in indices]

    # Return the reduced density matrix by using numpy tensor product
    density_matrix = np.tensordot(
        state, np.conj(state), axes=(traced_system, traced_system)
    )
    density_matrix = np.reshape(density_matrix, (2 ** len(indices), 2 ** len(indices)))

    return density_matrix


def reduced_dm(state, indices, check_state=False, c_dtype="complex128"):
    """Compute the reduced density matrix from a state vector or a density matrix. It supports all interfaces (Numpy,
    Autograd, Torch, Tensorflow and Jax).

    Args:
        state (tensor_like): ``(2**N)`` state vector or ``(2**N, 2**N)`` density matrix.
        indices (Sequence(int)): List of indices in the considered subsystem.
        check_state (bool): If True, the function will check the state validity (shape and norm).
        c_dtype (str): Complex floating point precision type.

    Returns:
        tensor_like: Reduced density matrix of size ``(2**len(indices), 2**len(indices))``

    **Example**

    >>> x = [1, 0, 1, 0] / np.sqrt(2)
    >>> reduced_dm(x, indices=[0])
    [[0.5+0.j 0.5+0.j]
     [0.5+0.j 0.5+0.j]]

    >>> reduced_dm(x, indices=[1])
    [[1.+0.j 0.+0.j]
     [0.+0.j 0.+0.j]]

    >>> y = tf.Variable([1, 0, 0, 0], dtype=tf.complex128)
    >>> reduced_dm(y, indices=[1])
    tf.Tensor(
    [[1.+0.j 0.+0.j]
     [0.+0.j 0.+0.j]], shape=(2, 2), dtype=complex128)

    >>> z = [[0.5, 0, 0.0, 0.5], [0, 0, 0, 0], [0, 0, 0, 0], [0.5, 0, 0, 0.5]]
    >>> reduced_dm(z, indices=[0])
    [[0.5+0.j 0.0+0.j]
     [0.0+0.j 0.5+0.j]]

    >>> reduced_dm(z, indices=[1])
    [[1.+0.j 0.+0.j]
     [0.+0.j 0.+0.j]]

    >>> y_mat_tf = tf.Variable([[1, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]], dtype=tf.complex128)
    >>> reduced_dm(y_mat_tf, indices=[1])
    tf.Tensor(
    [[1.+0.j 0.+0.j]
     [0.+0.j 0.+0.j]], shape=(2, 2), dtype=complex128)

    .. seealso:: :func:`pennylane.qinfo.transforms.reduced_dm` and :func:`pennylane.density_matrix`
    """
    # Cast as a c_dtype array
    state = cast(state, dtype=c_dtype)
    len_state = state.shape[0]
    # State vector
    if state.shape == (len_state,):
        density_matrix = _density_matrix_from_state_vector(state, indices, check_state)
        return density_matrix

    density_matrix = _density_matrix_from_matrix(state, indices, check_state)

    return density_matrix


def vn_entropy(state, indices, base=None, check_state=False, c_dtype="complex128"):
    r"""Compute the Von Neumann entropy from a state vector or density matrix on a given subsystem. It supports all
    interfaces (Numpy, Autograd, Torch, Tensorflow and Jax).

    .. math::
        S( \rho ) = -\text{Tr}( \rho \log ( \rho ))

    Args:
        state (tensor_like): ``(2**N)`` state vector or ``(2**N, 2**N)`` density matrix.
        indices (list(int)): List of indices in the considered subsystem.
        base (float): Base for the logarithm. If None, the natural logarithm is used.
        check_state (bool): If True, the function will check the state validity (shape and norm).
        c_dtype (str): Complex floating point precision type.

    Returns:
        float: Von Neumann entropy of the considered subsystem.

    **Example**

    The entropy of a subsystem for any state vectors can be obtained. Here is an example for the
    maximally entangled state, where the subsystem entropy is maximal (default base for log is exponential).


    >>> x = [1, 0, 0, 1] / np.sqrt(2)
    >>> vn_entropy(x, indices=[0])
    0.6931472

    The logarithm base can be switched to 2 for example.

    >>> vn_entropy(x, indices=[0], base=2)
    1.0

    The entropy can be obtained by providing a quantum state as a density matrix, for example:

    >>> y = [[1/2, 0, 0, 1/2], [0, 0, 0, 0], [0, 0, 0, 0], [1/2, 0, 0, 1/2]]
    >>> vn_entropy(x, indices=[0])
    0.6931472

    .. seealso:: :func:`pennylane.qinfo.transforms.vn_entropy` and :func:`pennylane.vn_entropy`
    """
    density_matrix = reduced_dm(state, indices, check_state, c_dtype)
    entropy = _compute_vn_entropy(density_matrix, base)

    return entropy


def _compute_vn_entropy(density_matrix, base=None):
    """Compute the Von Neumann entropy from a density matrix

    Args:
        density_matrix (tensor_like): ``(2**N, 2**N)`` tensor density matrix for an integer `N`.
        base (float, int): Base for the logarithm. If None, the natural logarithm is used.

    Returns:
        float: Von Neumann entropy of the density matrix.

    **Example**

    >>> x = [[1/2, 0], [0, 1/2]]
    >>> _compute_vn_entropy(x)
    0.6931472

    >>> x = [[1/2, 0], [0, 1/2]]
    >>> _compute_vn_entropy(x, base=2)
    1.0

    """
    # Change basis if necessary
    if base:
        div_base = np.log(base)
    else:
        div_base = 1

    evs = qml.math.eigvalsh(density_matrix)
    evs = qml.math.where(evs > 0, evs, 1.0)
    entropy = qml.math.entr(evs) / div_base

    return entropy


# pylint: disable=too-many-arguments
def mutual_info(
    state, indices0, indices1, base=None, check_state=False, c_dtype="complex128"
):
    r"""Compute the mutual information between two subsystems given a state:

    .. math::

        I(A, B) = S(\rho^A) + S(\rho^B) - S(\rho^{AB})

    where :math:`S` is the von Neumann entropy.

    The mutual information is a measure of correlation between two subsystems.
    More specifically, it quantifies the amount of information obtained about
    one system by measuring the other system. It supports all interfaces
    (Numpy, Autograd, Torch, Tensorflow and Jax).

    Each state can be given as a state vector in the computational basis, or
    as a density matrix.

    Args:
        state (tensor_like): ``(2**N)`` state vector or ``(2**N, 2**N)`` density matrix.
        indices0 (list[int]): List of indices in the first subsystem.
        indices1 (list[int]): List of indices in the second subsystem.
        base (float): Base for the logarithm. If None, the natural logarithm is used.
        check_state (bool): If True, the function will check the state validity (shape and norm).
        c_dtype (str): Complex floating point precision type.

    Returns:
        float: Mutual information between the subsystems

    **Examples**

    The mutual information between subsystems for a state vector can be returned as follows:

    >>> x = np.array([1, 0, 0, 1]) / np.sqrt(2)
    >>> qml.math.mutual_info(x, indices0=[0], indices1=[1])
    1.3862943611198906

    It is also possible to change the log basis.

    >>> qml.math.mutual_info(x, indices0=[0], indices1=[1], base=2)
    2.0

    Similarly the quantum state can be provided as a density matrix:

    >>> y = np.array([[1/2, 1/2, 0, 1/2], [1/2, 0, 0, 0], [0, 0, 0, 0], [1/2, 0, 0, 1/2]])
    >>> qml.math.mutual_info(y, indices0=[0], indices1=[1])
    0.4682351577408206

    .. seealso:: :func:`~.math.vn_entropy`, :func:`pennylane.qinfo.transforms.mutual_info` and :func:`pennylane.mutual_info`
    """

    # the subsystems cannot overlap
    if len([index for index in indices0 if index in indices1]) > 0:
        raise ValueError(
            "Subsystems for computing mutual information must not overlap."
        )

    # Cast to a complex array
    state = cast(state, dtype=c_dtype)

    state_shape = state.shape
    if len(state_shape) > 0:
        len_state = state_shape[0]
        if state_shape in [(len_state,), (len_state, len_state)]:
            return _compute_mutual_info(
                state,
                indices0,
                indices1,
                base=base,
                check_state=check_state,
                c_dtype=c_dtype,
            )

    raise ValueError("The state is not a state vector or a density matrix.")


# pylint: disable=too-many-arguments
def _compute_mutual_info(
    state, indices0, indices1, base=None, check_state=False, c_dtype="complex128"
):
    """Compute the mutual information between the subsystems."""
    all_indices = sorted([*indices0, *indices1])
    vn_entropy_1 = vn_entropy(
        state, indices=indices0, base=base, check_state=check_state, c_dtype=c_dtype
    )
    vn_entropy_2 = vn_entropy(
        state, indices=indices1, base=base, check_state=check_state, c_dtype=c_dtype
    )
    vn_entropy_12 = vn_entropy(
        state, indices=all_indices, base=base, check_state=check_state, c_dtype=c_dtype
    )

    return vn_entropy_1 + vn_entropy_2 - vn_entropy_12


def fidelity(state0, state1, check_state=False, c_dtype="complex128"):
    r"""Compute the fidelity for two states (a state can be a state vector or a density matrix) acting on quantum
    systems with the same size.

    The fidelity for two mixed states given by density matrices :math:`\rho` and :math:`\sigma`
    is defined as

    .. math::
        F( \rho , \sigma ) = \text{Tr}( \sqrt{\sqrt{\rho} \sigma \sqrt{\rho}})^2

    If one of the states is pure, say :math:`\rho=\ket{\psi}\bra{\psi}`, then the expression
    for fidelity simplifies to

    .. math::
        F( \ket{\psi} , \sigma ) = \bra{\psi} \sigma \ket{\psi}

    Finally, if both states are pure, :math:`\sigma=\ket{\phi}\bra{\phi}`, then the
    fidelity is simply

    .. math::
        F( \ket{\psi} , \ket{\phi}) = \left|\braket{\psi, \phi}\right|^2

    .. note::
        It supports all interfaces (Numpy, Autograd, Torch, Tensorflow and Jax). The second state is coerced
        to the type and dtype of the first state. The fidelity is returned in the type of the interface of the
        first state.

    Args:
        state0 (tensor_like): 1D state vector or 2D density matrix
        state1 (tensor_like): 1D state vector or 2D density matrix
        check_state (bool): If True, the function will check the validity of both states; it checks (shape, norm) for
            state vectors or (shape, trace, positive-definitiveness) for density matrices.
        c_dtype (str): Complex floating point precision type.

    Returns:
        float: Fidelity between the two quantum states.

    **Example**

    Two state vectors can be used as arguments and the fidelity (overlap) is returned, e.g.:

    >>> state0 = [0.98753537-0.14925137j, 0.00746879-0.04941796j]
    >>> state1 = [0.99500417+0.j, 0.09983342+0.j]
    >>> qml.math.fidelity(state0, state1)
    0.9905158135644924

    Alternatively one can give a state vector and a density matrix as arguments, e.g.:

    >>> state0 = [0, 1]
    >>> state1 = [[0, 0], [0, 1]]
    >>> qml.math.fidelity(state0, state1)
    1.0

    It also works with two density matrices, e.g.:

    >>> state0 = [[1, 0], [0, 0]]
    >>> state1 = [[0, 0], [0, 1]]
    >>> qml.math.fidelity(state0, state1)
    0.0

    .. seealso:: :func:`pennylane.qinfo.transforms.fidelity`

    """
    # Cast as a c_dtype array
    state0 = cast(state0, dtype=c_dtype)
    len_state0 = state0.shape[0]

    # Cannot be cast_like if jit
    if not is_abstract(state0):
        state1 = cast_like(state1, state0)

    len_state1 = state1.shape[0]

    if check_state:
        if state0.shape == (len_state0,):
            _check_state_vector(state0)
        else:
            _check_density_matrix(state0)

        if state1.shape == (len_state1,):
            _check_state_vector(state1)
        else:
            _check_density_matrix(state1)

    # Get dimension of the quantum system and reshape
    num_indices0 = int(np.log2(len_state0))
    num_indices1 = int(np.log2(len_state1))

    if num_indices0 != num_indices1:
        raise qml.QuantumFunctionError(
            "The two states must have the same number of wires."
        )

    # Two pure states, squared overlap
    if state1.shape == (len_state1,) and state0.shape == (len_state0,):
        overlap = np.tensordot(state0, np.transpose(np.conj(state1)), axes=1)
        overlap = np.linalg.norm(overlap) ** 2
        return overlap
    # First state mixed, second state pure
    if state1.shape == (len_state1,) and state0.shape != (len_state0,):
        overlap = np.tensordot(state0, np.transpose(np.conj(state1)), axes=1)
        overlap = np.tensordot(state1, overlap, axes=1)
        overlap = np.real(overlap)
        return overlap
    # First state pure, second state mixed
    if state0.shape == (len_state0,) and state1.shape != (len_state1,):
        overlap = np.tensordot(state1, np.transpose(np.conj(state0)), axes=1)
        overlap = np.tensordot(state0, overlap, axes=1)
        overlap = np.real(overlap)
        return overlap
    # Two mixed states
    fid = _compute_fidelity(state0, state1)
    return fid


def sqrt_matrix(density_matrix):
    r"""Compute the square root matrix of a density matrix where :math:`\rho = \sqrt{\rho} \times \sqrt{\rho}`
    Args:
        density_matrix (tensor_like): 2D density matrix of the quantum system.
    Returns:
        (tensor_like): Square root of the density matrix.
    """
    evs, vecs = qml.math.linalg.eigh(density_matrix)
    evs = np.real(evs)
    evs = qml.math.where(evs > 0.0, evs, 0.0)
    if not is_abstract(evs):
        evs = qml.math.cast_like(evs, vecs)
    return vecs @ qml.math.diag(np.sqrt(evs)) @ np.conj(np.transpose(vecs))


def _compute_fidelity(density_matrix0, density_matrix1):
    r"""Compute the fidelity for two density matrices with the same number of wires.

    .. math::
            F( \rho , \sigma ) = -\text{Tr}( \sqrt{\sqrt{\rho} \sigma \sqrt{\rho}})^2
    """
    # Implementation in single dispatches (sqrt(rho))
    sqrt_mat = qml.math.sqrt_matrix(density_matrix0)

    # sqrt(rho) * sigma * sqrt(rho)
    sqrt_mat_sqrt = sqrt_mat @ density_matrix1 @ sqrt_mat

    # extract eigenvalues
    evs = qml.math.eigvalsh(sqrt_mat_sqrt)
    evs = np.real(evs)
    evs = qml.math.where(evs > 0.0, evs, 0.0)

    trace = (qml.math.sum(qml.math.sqrt(evs))) ** 2

    return trace


def _compute_relative_entropy(rho, sigma, base=None):
    r"""
    Compute the quantum relative entropy of density matrix rho with respect to sigma.

    .. math::
        S(\rho\,\|\,\sigma)=-\text{Tr}(\rho\log\sigma)-S(\rho)=\text{Tr}(\rho\log\rho)-\text{Tr}(\rho\log\sigma)
        =\text{Tr}(\rho(\log\rho-\log\sigma))

    where :math:`S` is the von Neumann entropy.
    """
    if base:
        div_base = np.log(base)
    else:
        div_base = 1

    evs_rho, u_rho = qml.math.linalg.eigh(rho)
    evs_sig, u_sig = qml.math.linalg.eigh(sigma)

    # cast all eigenvalues to real
    evs_rho, evs_sig = np.real(evs_rho), np.real(evs_sig)

    # zero eigenvalues need to be treated very carefully here
    # we use the convention that 0 * log(0) = 0
    evs_sig = qml.math.where(evs_sig == 0, 0.0, evs_sig)
    rho_nonzero_mask = qml.math.where(evs_rho == 0.0, False, True)

    ent = qml.math.entr(qml.math.where(rho_nonzero_mask, evs_rho, 1.0))

    # the matrix of inner products between eigenvectors of rho and eigenvectors
    # of sigma; this is a doubly stochastic matrix
    rel = np.abs(qml.math.dot(np.transpose(np.conj(u_rho)), u_sig)) ** 2

    rel = qml.math.sum(qml.math.where(rel == 0.0, 0.0, np.log(evs_sig) * rel), axis=1)
    rel = -qml.math.sum(qml.math.where(rho_nonzero_mask, evs_rho * rel, 0.0))

    return (rel - ent) / div_base


def relative_entropy(
    state0, state1, base=None, check_state=False, c_dtype="complex128"
):
    r"""
    Compute the quantum relative entropy of one state with respect to another.

    .. math::
        S(\rho\,\|\,\sigma)=-\text{Tr}(\rho\log\sigma)-S(\rho)=\text{Tr}(\rho\log\rho)-\text{Tr}(\rho\log\sigma)
        =\text{Tr}(\rho(\log\rho-\log\sigma))

    Roughly speaking, quantum relative entropy is a measure of distinguishability between two
    quantum states. It is the quantum mechanical analog of relative entropy.

    Each state can be given as a state vector in the computational basis or
    as a density matrix.

    Args:
        state0 (tensor_like): ``(2**N)`` state vector or ``(2**N, 2**N)`` density matrix.
        state1 (tensor_like): ``(2**N)`` state vector or ``(2**N, 2**N)`` density matrix.
        base (float): Base for the logarithm. If None, the natural logarithm is used.
        check_state (bool): If True, the function will check the state validity (shape and norm).
        c_dtype (str): Complex floating point precision type.

    Returns:
        float: Quantum relative entropy of state0 with respect to state1

    **Examples**

    The relative entropy between two equal states is always zero:

    >>> x = np.array([1, 0])
    >>> qml.math.relative_entropy(x, x)
    0.0

    and the relative entropy between two non-equal pure states is always infinity:

    >>> y = np.array([1, 1]) / np.sqrt(2)
    >>> qml.math.relative_entropy(x, y)
    inf

    The quantum states can be provided as density matrices, allowing for computation
    of relative entropy between mixed states:

    >>> rho = np.array([[0.3, 0], [0, 0.7]])
    >>> sigma = np.array([[0.5, 0], [0, 0.5]])
    >>> qml.math.relative_entropy(rho, sigma)
    tensor(0.08228288, requires_grad=True)

    It is also possible to change the log base:

    >>> qml.math.relative_entropy(rho, sigma, base=2)
    tensor(0.1187091, requires_grad=True)

    .. seealso:: :func:`pennylane.qinfo.transforms.relative_entropy`
    """
    # Cast as a c_dtype array
    state0 = cast(state0, dtype=c_dtype)
    len_state0 = state0.shape[0]

    # Cannot be cast_like if jit
    if not is_abstract(state0):
        state1 = cast_like(state1, state0)

    len_state1 = state1.shape[0]

    if check_state:
        if state0.shape == (len_state0,):
            _check_state_vector(state0)
        else:
            _check_density_matrix(state0)

        if state1.shape == (len_state1,):
            _check_state_vector(state1)
        else:
            _check_density_matrix(state1)

    # Get dimension of the quantum system and reshape
    num_indices0 = int(np.log2(len_state0))
    num_indices1 = int(np.log2(len_state1))

    if num_indices0 != num_indices1:
        raise qml.QuantumFunctionError(
            "The two states must have the same number of wires."
        )

    if state0.shape == (len_state0,):
        state0 = qml.math.outer(state0, np.conj(state0))

    if state1.shape == (len_state1,):
        state1 = qml.math.outer(state1, np.conj(state1))

    return _compute_relative_entropy(state0, state1, base=base)


def _check_density_matrix(density_matrix):
    """Check the shape, the trace and the positive semi-definitiveness of a matrix."""
    shape = density_matrix.shape[0]
    if (
        len(density_matrix.shape) != 2
        or density_matrix.shape[0] != density_matrix.shape[1]
        or not np.log2(shape).is_integer()
    ):
        raise ValueError("Density matrix must be of shape (2**N, 2**N).")
    # Check trace
    trace = np.trace(density_matrix)
    if not is_abstract(trace):
        if not allclose(trace, 1.0, atol=1e-10):
            raise ValueError("The trace of the density matrix should be one.")
        # Check if the matrix is Hermitian
        conj_trans = np.transpose(np.conj(density_matrix))
        if not allclose(density_matrix, conj_trans):
            raise ValueError("The matrix is not Hermitian.")
        # Check if positive semi-definite
        evs = np.linalg.eigvalsh(density_matrix)
        evs = np.real(evs)
        evs_non_negative = [ev for ev in evs if ev >= 0.0]
        if len(evs) != len(evs_non_negative):
            raise ValueError("The matrix is not positive semi-definite.")


def _check_state_vector(state_vector):
    """Check the shape and the norm of a state vector."""
    len_state = state_vector.shape[0]
    # Check format
    if len(np.shape(state_vector)) != 1 or not np.log2(len_state).is_integer():
        raise ValueError("State vector must be of length 2**wires.")
    # Check norm
    norm = np.linalg.norm(state_vector, ord=2)
    if not is_abstract(norm):
        if not allclose(norm, 1.0, atol=1e-10):
            raise ValueError("Sum of amplitudes-squared does not equal one.")



from .backends import DeviceAzure

from openqaoa_qiskit.backends import QAOAQiskitQPUBackend

device_access = {DeviceAzure: QAOAQiskitQPUBackend}
device_location = "azure"
device_plugin = DeviceAzure
device_args = {DeviceAzure: ["resource_id", "location"]}
